# Third-party imports
import pytest

# Local imports
from uplink.clients.io import interfaces, state, transitions


@pytest.fixture
def request_execution_mock(mocker):
    return mocker.Mock(spec=interfaces.RequestExecution)


@pytest.fixture
def request_state_mock(mocker):
    return mocker.Mock(spec=interfaces.RequestState)


class BasicStateTest(object):
    def create_state(self, request):
        raise NotImplementedError

    @staticmethod
    def _create_request_mock():
        return object()

    def test_prepare(self):
        request = self._create_request_mock()
        target = self.create_state(request)
        output = target.prepare(request)
        assert output == state.BeforeRequest(request)

    def test_send(self):
        request = self._create_request_mock()
        target = self.create_state(request)
        output = target.send(request)
        assert output == state.SendRequest(request)

    def test_fail(self):
        request = self._create_request_mock()
        target = self.create_state(request)
        error = Exception()
        output = target.fail(Exception, error, None)
        assert output == state.Fail(request, Exception, error, None)

    def test_finish(self):
        request = self._create_request_mock()
        response = object()
        target = self.create_state(request)
        output = target.finish(response)
        assert output == state.Finish(request, response)

    def test_sleep(self):
        request = self._create_request_mock()
        target = self.create_state(request)
        output = target.sleep(10)
        assert output == state.Sleep(request, 10)

    def test_request_property(self):
        request = self._create_request_mock()
        target = self.create_state(request)
        assert target.request == request


class TestBeforeRequest(BasicStateTest):
    def create_state(self, request):
        return state.BeforeRequest(request)


class TestAfterResponse(BasicStateTest):
    def create_state(self, request):
        return state.AfterResponse(request, object())


class TestAfterException(BasicStateTest):
    def create_state(self, request):
        return state.AfterException(request, Exception, Exception(), None)


class TestSleep(object):
    def test_execute(self, request_execution_mock):
        request = object()
        sleep = state.Sleep(request, 10)
        sleep.execute(request_execution_mock)
        assert request_execution_mock.sleep.called

        args, _ = request_execution_mock.sleep.call_args
        callback = args[1]
        assert isinstance(callback, interfaces.SleepCallback)

        callback.on_success()
        assert request_execution_mock.state == state.BeforeRequest(request)

        error = Exception()
        callback.on_failure(Exception, error, None)
        assert request_execution_mock.state == state.AfterException(
            request, Exception, error, None
        )


class TestSendRequest(object):
    def test_execute(self, request_execution_mock):
        request = object()
        send_request = state.SendRequest(request)
        send_request.execute(request_execution_mock)
        assert request_execution_mock.send.called

        args, _ = request_execution_mock.send.call_args
        callback = args[1]
        assert isinstance(callback, interfaces.InvokeCallback)

        response = object()
        callback.on_success(response)
        assert request_execution_mock.state == state.AfterResponse(
            request, response
        )

        error = Exception()
        callback.on_failure(Exception, error, None)
        assert request_execution_mock.state == state.AfterException(
            request, Exception, error, None
        )


class TestFail(object):
    def test_execute(self, request_execution_mock):
        request, error = object(), Exception()
        fail = state.Fail(request, type(error), error, None)
        fail.execute(request_execution_mock)
        request_execution_mock.fail.assert_called_with(Exception, error, None)


class TestFinish(object):
    def test_execute(self, request_execution_mock):
        request, response = object(), object()
        finish = state.Finish(request, response)
        finish.execute(request_execution_mock)
        request_execution_mock.finish.assert_called_with(response)


def test_sleep_transition(request_state_mock):
    transitions.sleep(10)(request_state_mock)
    request_state_mock.sleep.assert_called_with(10)


def test_send_transition(request_state_mock):
    request = object()
    transitions.send(request)(request_state_mock)
    request_state_mock.send.assert_called_with(request)


def test_finish_transition(request_state_mock):
    response = object()
    transitions.finish(response)(request_state_mock)
    request_state_mock.finish.assert_called_with(response)


def test_fail_transition(request_state_mock):
    error = Exception()
    transitions.fail(Exception, error, None)(request_state_mock)
    request_state_mock.fail.assert_called_with(Exception, error, None)


def test_prepare_transition(request_state_mock):
    request = object()
    transitions.prepare(request)(request_state_mock)
    request_state_mock.prepare.assert_called_with(request)



# Third-party imports
import pytest

# Local imports
from uplink import returns
from uplink.decorators import json
from uplink.models import loads, dumps


@pytest.mark.parametrize(
    "cls, method",
    [
        (loads, "create_response_body_converter"),
        (dumps, "create_request_body_converter"),
    ],
)
def test_models(mocker, cls, method, request_definition):
    # Setup
    func = mocker.stub()
    func.return_value = 1

    # Verify: Returns obj that wraps func
    obj = cls(object, (object,))
    factory = obj(func)
    assert callable(factory)
    assert factory(1) == 1

    # Verify: not relevant
    request_definition.argument_annotations = ()
    request_definition.method_annotations = ()
    value = getattr(factory, method)(None, request_definition)
    assert value is None

    # Verify: relevant
    request_definition.argument_annotations = (object(),)
    value = getattr(factory, method)(object, request_definition)
    assert callable(value)


@pytest.mark.parametrize(
    "cls, method, decorator",
    [
        (loads.from_json, "create_response_body_converter", returns.json()),
        (dumps.to_json, "create_request_body_converter", json()),
    ],
)
def test_json_builders(mocker, cls, method, decorator, request_definition):
    # Setup
    func = mocker.stub()
    func.return_value = 1
    obj = cls(object)
    factory = obj.using(func)

    # Verify: not relevant
    request_definition.argument_annotations = ()
    request_definition.method_annotations = ()
    value = getattr(factory, method)(object, request_definition)
    assert value is None

    # Verify relevant
    request_definition.method_annotations = (decorator,)
    value = getattr(factory, method)(object, request_definition)
    assert callable(value)
    assert value(1) == 1



# Local imports
from uplink import arguments


class Session(object):
    """
    The session of a :class:`~uplink.Consumer` instance.

    Exposes the configuration of a :class:`~uplink.Consumer` instance and
    allows the persistence of certain properties across all requests sent
    from that instance.
    """

    def __init__(self, builder):
        self.__builder = builder
        self.__params = None
        self.__headers = None
        self.__context = None

    def create(self, consumer, definition):
        return self.__builder.build(definition, consumer)

    @property
    def base_url(self):
        """
        The base URL for any requests sent from this consumer instance.
        """
        return self.__builder.base_url

    @property
    def headers(self):
        """
        A dictionary of headers to be sent on each request from this
        consumer instance.
        """
        if self.__headers is None:
            self.__headers = {}
            self.inject(arguments.HeaderMap().with_value(self.__headers))
        return self.__headers

    @property
    def params(self):
        """
        A dictionary of querystring data to attach to each request from
        this consumer instance.
        """
        if self.__params is None:
            self.__params = {}
            self.inject(arguments.QueryMap().with_value(self.__params))
        return self.__params

    @property
    def context(self):
        """
        A dictionary of name-value pairs that are made available to
        request middleware.
        """
        if self.__context is None:
            self.__context = {}
            self.inject(arguments.ContextMap().with_value(self.__context))
        return self.__context

    @property
    def auth(self):
        """The authentication object for this consumer instance."""
        return self.__builder.auth

    @auth.setter
    def auth(self, auth):
        self.__builder.auth = auth

    def inject(self, hook, *more_hooks):
        """
        Add hooks (e.g., functions decorated with either
        :class:`~uplink.response_handler` or
        :class:`~uplink.error_handler`) to the session.
        """
        self.__builder.add_hook(hook, *more_hooks)



# Third-party imports
import pytest

# Local imports.
import uplink

# Constants
BASE_URL = "https://api.github.com/"


@uplink.timeout(10)
@uplink.headers({"Accept": "application/vnd.github.v3.full+json"})
class GitHubService(uplink.Consumer):
    @uplink.timeout(15)
    @uplink.get("/users/{user}/repos")
    def list_repos(self, user):
        """List all public repositories for a specific user."""

    @uplink.returns.json
    @uplink.get("/users/{user}/repos/{repo}")
    def get_repo(self, user, repo):
        pass

    @uplink.get(args={"url": uplink.Url})
    def forward(self, url):
        pass


def test_list_repo_wrapper(mock_client):
    """Ensures that the consumer method looks like the original func."""
    github = GitHubService(base_url=BASE_URL, client=mock_client)
    assert (
        github.list_repos.__doc__
        == GitHubService.list_repos.__doc__
        == "List all public repositories for a specific user."
    )
    assert (
        github.list_repos.__name__
        == GitHubService.list_repos.__name__
        == "list_repos"
    )


def test_list_repo(mock_client):
    github = GitHubService(base_url=BASE_URL, client=mock_client)
    github.list_repos("prkumar")
    request = mock_client.history[0]
    assert request.method == "GET"
    assert request.has_base_url(BASE_URL)
    assert request.has_endpoint("/users/prkumar/repos")
    assert request.headers == {"Accept": "application/vnd.github.v3.full+json"}
    assert request.timeout == 15


def test_get_repo(mock_client, mock_response):
    """
    This integration test ensures that the returns.json returns the
    Json body when a model is not provided.
    """
    # Setup: return a mock response
    expected_json = {"key": "value"}
    mock_response.with_json(expected_json)
    mock_client.with_response(mock_response)
    github = GitHubService(base_url=BASE_URL, client=mock_client)

    # Run
    actual_json = github.get_repo("prkumar", "uplink")

    # Verify
    assert expected_json == actual_json


def test_forward(mock_client):
    github = GitHubService(base_url=BASE_URL, client=mock_client)
    github.forward("/users/prkumar/repos")
    request = mock_client.history[0]
    assert request.method == "GET"
    assert request.has_base_url(BASE_URL)
    assert request.has_endpoint("/users/prkumar/repos")


def test_handle_client_exceptions(mock_client):
    # Setup: mock client exceptions

    class MockBaseClientException(Exception):
        pass

    class MockInvalidURL(MockBaseClientException):
        pass

    mock_client.exceptions.BaseClientException = MockBaseClientException
    mock_client.exceptions.InvalidURL = MockInvalidURL

    # Setup: instantiate service
    service = GitHubService(base_url=BASE_URL, client=mock_client)

    # Run: Catch base exception
    mock_client.with_side_effect(MockBaseClientException)

    with pytest.raises(service.exceptions.BaseClientException):
        service.list_repos("prkumar")

    # Run: Catch leaf exception
    mock_client.with_side_effect(MockInvalidURL)

    with pytest.raises(service.exceptions.InvalidURL):
        service.list_repos("prkumar")

    # Run: Try polymorphism
    mock_client.with_side_effect(MockInvalidURL)

    with pytest.raises(service.exceptions.BaseClientException):
        service.list_repos("prkumar")



# Local imports
from uplink.clients.io import interfaces
from uplink.clients.io import state

__all__ = ["RequestExecutionBuilder"]


class RequestExecutionBuilder(object):
    def __init__(self):
        self._client = None
        self._template = None
        self._io = None
        self._callbacks = []
        self._errbacks = []

    def with_client(self, client):
        self._client = client
        return self

    def with_template(self, template):
        self._template = template
        return self

    def with_io(self, io):
        self._io = io
        return self

    def with_callbacks(self, *callbacks):
        self._callbacks.extend(callbacks)
        return self

    def with_errbacks(self, *errbacks):
        self._errbacks.extend(errbacks)
        return self

    def build(self):
        client, io = self._client, self._io
        for callback in self._callbacks:
            io = CallbackDecorator(io, client, callback)
        for errback in self._errbacks:
            io = ErrbackDecorator(io, errback)
        return DefaultRequestExecution(client, io, self._template)


class DefaultRequestExecution(interfaces.RequestExecution):
    def __init__(self, client, io, template):
        self._client = client
        self._template = template
        self._io = io
        self._state = None

    def before_request(self, request):
        action = self._template.before_request(request)
        next_state = action(self._state)
        self._state = next_state
        return self.execute()

    def after_response(self, request, response):
        action = self._template.after_response(request, response)
        next_state = action(self._state)
        self._state = next_state
        return self.execute()

    def after_exception(self, request, exc_type, exc_val, exc_tb):
        action = self._template.after_exception(
            request, exc_type, exc_val, exc_tb
        )
        next_state = action(self._state)
        self._state = next_state
        return self.execute()

    def send(self, request, callback):
        return self._io.invoke(self._client.send, (request,), {}, callback)

    def sleep(self, duration, callback):
        return self._io.sleep(duration, callback)

    def finish(self, response):
        return self._io.finish(response)

    def fail(self, exc_type, exc_val, exc_tb):
        return self._io.fail(exc_type, exc_val, exc_tb)

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, new_state):
        self._state = new_state

    def execute(self):
        return self.state.execute(self)

    def start(self, request):
        self._state = state.BeforeRequest(request)  # Start state
        return self._io.execute(self)


class FinishingCallback(interfaces.InvokeCallback):
    def __init__(self, io):
        self._io = io

    def on_success(self, result):
        return self._io.finish(result)

    def on_failure(self, exc_type, exc_val, exc_tb):
        return self._io.fail(exc_type, exc_val, exc_tb)


class IOStrategyDecorator(interfaces.IOStrategy):
    def __init__(self, io):
        self._io = io

    def invoke(self, func, args, kwargs, callback):
        return self._io.invoke(func, args, kwargs, callback)

    def sleep(self, duration, callback):
        return self._io.sleep(duration, callback)

    def execute(self, executable):
        return self._io.execute(executable)

    def finish(self, response):
        return self._io.finish(response)

    def fail(self, exc_type, exc_val, exc_tb):  # pragma: no cover
        return self._io.fail(exc_type, exc_val, exc_tb)


class FinishingDecorator(IOStrategyDecorator):
    def _invoke(self, func, *args, **kwargs):
        return self._io.invoke(func, args, kwargs, FinishingCallback(self._io))


class CallbackDecorator(FinishingDecorator):
    def __init__(self, io, client, callback):
        super(CallbackDecorator, self).__init__(io)
        self._client = client
        self._callback = callback

    def finish(self, response):
        return self._invoke(
            self._client.apply_callback, self._callback, response
        )


class ErrbackDecorator(FinishingDecorator):
    def __init__(self, io, errback):
        super(ErrbackDecorator, self).__init__(io)
        self._errback = errback

    def fail(self, exc_type, exc_val, exc_tb):
        return self._invoke(self._errback, exc_type, exc_val, exc_tb)



# Standard library imports
import collections

# Local imports
from uplink import interfaces, utils
from uplink.clients import io


def get_api_definitions(service):
    """
    Returns all attributes with type
    `uplink.interfaces.RequestDefinitionBuilder` defined on the given
    class.

    Note:
        All attributes are considered, not only defined directly on the class.

    Args:
        service: A class object.
    """
    # In Python 3.3, `inspect.getmembers` doesn't respect the descriptor
    # protocol when the first argument is a class. In other words, the
    # function includes any descriptors bound to `service` as is rather
    # than calling the descriptor's __get__ method. This is seemingly
    # fixed in Python 2.7 and 3.4+ (TODO: locate corresponding bug
    # report in Python issue tracker). Directly invoking `getattr` to
    # force Python's attribute lookup protocol is a decent workaround to
    # ensure parity:
    class_attributes = ((k, getattr(service, k)) for k in dir(service))

    is_definition = interfaces.RequestDefinitionBuilder.__instancecheck__
    return [(k, v) for k, v in class_attributes if is_definition(v)]


def set_api_definition(service, name, definition):
    setattr(service, name, definition)


class RequestBuilder(object):
    def __init__(self, client, converter_registry, base_url):
        self._method = None
        self._relative_url_template = utils.URIBuilder("")
        self._return_type = None
        self._client = client
        self._base_url = base_url

        # TODO: Pass this in as constructor parameter
        # TODO: Delegate instantiations to uplink.HTTPClientAdapter
        self._info = collections.defaultdict(dict)
        self._context = {}

        self._converter_registry = converter_registry
        self._transaction_hooks = []
        self._request_templates = []

    @property
    def client(self):
        return self._client

    @property
    def method(self):
        return self._method

    @method.setter
    def method(self, method):
        self._method = method

    @property
    def base_url(self):
        return self._base_url

    def set_url_variable(self, variables):
        self._relative_url_template.set_variable(variables)

    @property
    def relative_url(self):
        return self._relative_url_template.build()

    @relative_url.setter
    def relative_url(self, url):
        self._relative_url_template = utils.URIBuilder(url)

    @property
    def info(self):
        return self._info

    @property
    def context(self):
        return self._context

    @property
    def transaction_hooks(self):
        return iter(self._transaction_hooks)

    def get_converter(self, converter_key, *args, **kwargs):
        return self._converter_registry[converter_key](*args, **kwargs)

    @property
    def return_type(self):
        return self._return_type

    @return_type.setter
    def return_type(self, return_type):
        self._return_type = return_type

    @property
    def request_template(self):
        return io.CompositeRequestTemplate(self._request_templates)

    @property
    def url(self):
        return utils.urlparse.urljoin(self.base_url, self.relative_url)

    def add_transaction_hook(self, hook):
        self._transaction_hooks.append(hook)

    def add_request_template(self, template):
        self._request_templates.append(template)



class Converter(object):
    def convert(self, value):
        raise NotImplementedError

    def __call__(self, *args, **kwargs):
        return self.convert(*args, **kwargs)

    def set_chain(self, chain):
        pass


class Factory(object):
    """
    An adapter that handles serialization of HTTP request properties
    (e.g., headers, query parameters, request body) and deserialization
    of HTTP response bodies.

    Each concrete implementation of this abstract class typically
    encapsulates a specific encoding/decoding strategy
    (e.g., Protocol Buffers or JSON).

    .. note::

        Overriding all inherited methods is unnecessary; the default
        implementation is to return :obj:`None`, which tells the
        converter layer to move on to the next factory. Hence,
        you only should implement the methods you intend to support.

    """

    def create_response_body_converter(self, cls, request_definition):
        """
        Returns a callable that can convert a response body into the
        specified :obj:`cls`.

        The returned callable should expect a single positional
        argument: the response body.

        If this factory can't produce such a callable, it should return
        :obj:`None`, so another factory can have a chance to handle
        the type.

        Args:
            cls (:obj:`type`): The target class for conversion.
            request_definition: Metadata for the outgoing request.
                This object exposes two properties: the
                :obj:`method_annotations` (e.g., `~uplink.headers`) and
                :obj:`argument_annotations` (e.g., `~uplink.Body) bound
                to the underlying consumer method
        """

    def create_request_body_converter(self, cls, request_definition):
        """
        Returns a callable that can convert :obj:`cls` into an acceptable
        request body.

        The returned callable should expect a single positional
        argument: an instance of given type, :obj:`cls`.

        If this factory can't produce such a callable, it should return
        :py:obj:`None`, so another factory can have a chance to handle
        the type.

        Args:
            cls (obj:`type`): The target class for conversion.
            request_definition: Metadata for the outgoing request.
                This object exposes two properties: the
                :obj:`method_annotations` (e.g., `~uplink.headers`) and
                :obj:`argument_annotations` (e.g., `~uplink.Body) bound
                to the underlying consumer method
        """

    def create_string_converter(self, cls, request_definition):
        """
        Returns a callable that can convert `cls` into a
        :py:class:`str`.

        The returned callable should expect a single positional
        argument: an instance of given type, :obj:`cls`.

        If this factory can't produce such a callable, it should return
        :py:obj:`None`, so another factory can have a chance to handle
        the type.

        Args:
            cls (obj:`type`): The target class for conversion.
            request_definition: Metadata for the outgoing request.
                This object exposes two properties: the
                :obj:`method_annotations` (e.g., `~uplink.headers`) and
                :obj:`argument_annotations` (e.g., `~uplink.Body) bound
                to the underlying consumer method
        """


class ConverterFactory(Factory):
    # TODO: Remove this in v1.0.0 -- use Factory instead.

    def create_response_body_converter(self, cls, request_definition):
        return self.make_response_body_converter(
            cls,
            request_definition.argument_annotations,
            request_definition.method_annotations,
        )

    def create_request_body_converter(self, cls, request_definition):
        return self.make_request_body_converter(
            cls,
            request_definition.argument_annotations,
            request_definition.method_annotations,
        )

    def create_string_converter(self, cls, request_definition):
        return self.make_string_converter(
            cls,
            request_definition.argument_annotations,
            request_definition.method_annotations,
        )

    def make_response_body_converter(
        self, type, argument_annotations, method_annotations
    ):
        pass

    def make_request_body_converter(
        self, type, argument_annotations, method_annotations
    ):
        pass

    def make_string_converter(
        self, type, argument_annotations, method_annotations
    ):
        pass



# Standard library imports
import typing
import sys

# Third-party imports
import marshmallow

try:
    import pydantic
except ImportError:
    if sys.version_info >= (3, 6):
        raise

import pytest

# Local imports
from uplink import converters
from uplink.converters import register, standard


class TestStringConverter(object):
    def test_convert(self):
        converter_ = standard.StringConverter()
        assert converter_.convert(2) == "2"


class TestStandardConverter(object):
    def test_create_response_body_converter_with_converter(
        self, converter_mock
    ):
        # Setup
        factory = standard.StandardConverter()

        # Run & Verify: Pass-through converters
        converter = factory.create_response_body_converter(converter_mock)
        assert converter is converter_mock

    def test_create_response_body_converter_with_unknown_type(self):
        # Setup
        factory = standard.StandardConverter()

        # Run & Verify: does not know how to create converter when given a type
        # that is not a converter
        converter = factory.create_response_body_converter(dict)
        assert converter is None

    def test_create_request_body_converter_with_converter(self, converter_mock):
        # Setup
        factory = standard.StandardConverter()

        # Run & Verify: Pass-through converters
        converter = factory.create_request_body_converter(converter_mock)
        assert converter is converter_mock

    def test_create_request_body_converter_with_unknown_type(self):
        # Setup
        factory = standard.StandardConverter()

        # Run & Verify: does not know how to create converter when given a type
        # that is not a converter
        converter = factory.create_response_body_converter(dict)
        assert converter is None

    def test_create_string_converter_with_converter(self, converter_mock):
        # Setup
        factory = standard.StandardConverter()

        # Run & Verify: Pass-through converters
        converter = factory.create_string_converter(converter_mock)
        assert converter is converter_mock

    def test_create_string_converter_with_unknown_type(self):
        # Setup
        factory = standard.StandardConverter()

        # Run & Verify: creates string converter when given type is not a
        # converter
        converter = factory.create_string_converter(dict)
        assert isinstance(converter, standard.StringConverter)


class TestConverterFactoryRegistry(object):
    backend = converters.ConverterFactoryRegistry._converter_factory_registry

    def test_init_args_are_passed_to_factory(
        self, converter_factory_mock, converter_mock
    ):
        args = ("arg1", "arg2")
        kwargs = {"arg3": "arg3"}
        converter_factory_mock.create_string_converter.return_value = (
            converter_mock
        )
        registry = converters.ConverterFactoryRegistry(
            (converter_factory_mock,), *args, **kwargs
        )
        return_value = registry[converters.keys.CONVERT_TO_STRING]()
        converter_factory_mock.create_string_converter.assert_called_with(
            *args, **kwargs
        )
        assert return_value is converter_mock

        # Test with type that can't be handled by registry
        converter_factory_mock.create_string_converter.return_value = None
        return_value = registry[converters.keys.CONVERT_TO_STRING]()
        assert return_value is None

    def test_hooks(self, converter_factory_mock, converter_mock):
        converter_factory_mock.create_string_converter.return_value = (
            converter_mock
        )
        registry = converters.ConverterFactoryRegistry(
            (converter_factory_mock,)
        )
        registry[converters.keys.CONVERT_TO_STRING]()
        assert converter_mock.set_chain.called

    def test_len(self):
        registry = converters.ConverterFactoryRegistry(())
        assert len(registry) == len(self.backend)

    def test_iter(self):
        registry = converters.ConverterFactoryRegistry(())
        assert list(iter(registry)) == list(iter(self.backend))


def test_create_request_body_converter(converter_factory_mock):
    method = converters.create_request_body_converter(converter_factory_mock)
    assert method is converter_factory_mock.create_request_body_converter


def test_create_response_body_converter(converter_factory_mock):
    method = converters.create_response_body_converter(converter_factory_mock)
    assert method is converter_factory_mock.create_response_body_converter


def test_create_string_converter(converter_factory_mock):
    method = converters.create_string_converter(converter_factory_mock)
    assert method is converter_factory_mock.create_string_converter


@pytest.fixture(params=["class", "instance"])
def schema_mock_and_argument(request, mocker):
    class Schema(marshmallow.Schema):
        def __new__(cls, *args, **kwargs):
            return schema

    schema = mocker.Mock(spec=marshmallow.Schema)
    if request.param == "class":
        return schema, Schema
    else:
        return schema, schema


class TestMarshmallowConverter(object):
    for_marshmallow_2_and_3 = pytest.mark.parametrize(
        "is_marshmallow_3", [True, False]
    )

    @staticmethod
    def _mock_data(mocker, expected_result, is_marshmallow_3):
        """Mocks the result of Schema.dump() or Schema.load()"""

        if is_marshmallow_3:
            # After marshmallow 3.0, Schema.load() and Schema.dump() don't
            # return a (data, errors) tuple any more. Only `data` is returned.
            return expected_result

        result = mocker.Mock()
        result.data = expected_result
        return result

    def test_init_without_marshmallow(self):
        old_marshmallow = converters.MarshmallowConverter.marshmallow
        converters.MarshmallowConverter.marshmallow = None
        with pytest.raises(ImportError):
            converters.MarshmallowConverter()
        converters.MarshmallowConverter.marshmallow = old_marshmallow

    @for_marshmallow_2_and_3
    def test_create_request_body_converter(
        self, mocker, schema_mock_and_argument, is_marshmallow_3
    ):
        # Setup
        schema_mock, argument = schema_mock_and_argument
        expected_result = "data"
        schema_mock.dump.return_value = self._mock_data(
            mocker, expected_result, is_marshmallow_3
        )
        converter = converters.MarshmallowConverter()
        converter.is_marshmallow_3 = is_marshmallow_3
        request_body = {"id": 0}

        # Run
        c = converter.create_request_body_converter(argument)
        result = c.convert(request_body)

        # Verify
        schema_mock.dump.assert_called_with(request_body)
        assert expected_result == result

    def test_create_request_body_converter_without_schema(self):
        # Setup
        converter = converters.MarshmallowConverter()

        # Run
        c = converter.create_request_body_converter("not a schema")

        # Verify
        assert c is None

    @for_marshmallow_2_and_3
    def test_create_response_body_converter(
        self, mocker, schema_mock_and_argument, is_marshmallow_3
    ):
        # Setup
        schema_mock, argument = schema_mock_and_argument
        expected_result = "data"
        schema_mock.load.return_value = self._mock_data(
            mocker, expected_result, is_marshmallow_3
        )
        converter = converters.MarshmallowConverter()
        converter.is_marshmallow_3 = is_marshmallow_3
        response = mocker.Mock(spec=["json"])
        c = converter.create_response_body_converter(argument)

        # Run & Verify: with response
        result = c.convert(response)
        response.json.assert_called_with()
        schema_mock.load.assert_called_with(response.json())
        assert expected_result == result

        # Run & Verify: with json
        data = {"hello": "world"}
        result = c.convert(data)
        schema_mock.load.assert_called_with(data)
        assert expected_result == result

        # Run & Verify: raise validation errors for user to handle.
        schema_mock.load.side_effect = marshmallow.exceptions.MarshmallowError

        with pytest.raises(marshmallow.exceptions.MarshmallowError):
            c.convert(data)

    def test_create_response_body_converter_with_unsupported_response(
        self, schema_mock_and_argument
    ):
        # Setup
        schema_mock, argument = schema_mock_and_argument
        converter = converters.MarshmallowConverter()

        # Run
        c = converter.create_response_body_converter(argument)
        result = c.convert("unsupported response")

        # Verify
        return result is None

    def test_create_response_body_converter_without_schema(self):
        # Setup
        converter = converters.MarshmallowConverter()

        # Run
        c = converter.create_response_body_converter("not a schema")

        # Verify
        assert c is None

    def test_create_string_converter(self, schema_mock_and_argument):
        # Setup
        _, argument = schema_mock_and_argument
        converter = converters.MarshmallowConverter()

        # Run
        c = converter.create_string_converter(argument, None)

        # Verify
        assert c is None

    def test_register(self):
        # Setup
        converter = converters.MarshmallowConverter
        old_marshmallow = converter.marshmallow

        # Run & Verify: Register when marshmallow is installed
        converter.marshmallow = True
        register_ = []
        converter.register_if_necessary(register_.append)
        assert register_ == [converter]

        # Run & Verify: Skip when marshmallow is not installed
        converter.marshmallow = None
        register_ = []
        converter.register_if_necessary(register_.append)
        assert register_ == []

        # Rewire
        converters.MarshmallowConverter.marshmallow = old_marshmallow


class TestMap(object):
    def test_convert(self):
        # Setup
        registry = converters.ConverterFactoryRegistry(
            (converters.StandardConverter(),)
        )
        key = converters.keys.Map(converters.keys.CONVERT_TO_STRING)

        # Run
        converter = registry[key](None)
        value = converter({"hello": 1})

        # Verify
        assert value == {"hello": "1"}

    def test_eq(self):
        assert converters.keys.Map(0) == converters.keys.Map(0)
        assert not (converters.keys.Map(1) == converters.keys.Map(0))
        assert not (converters.keys.Map(1) == 1)


class TestSequence(object):
    def test_convert_with_sequence(self):
        # Setup
        registry = converters.ConverterFactoryRegistry(
            (converters.StandardConverter(),)
        )
        key = converters.keys.Sequence(converters.keys.CONVERT_TO_STRING)

        # Run
        converter = registry[key](None)
        value = converter([1, 2, 3])

        # Verify
        assert value == ["1", "2", "3"]

    def test_convert_not_sequence(self):
        # Setup
        registry = converters.ConverterFactoryRegistry(
            (converters.StandardConverter(),)
        )
        key = converters.keys.Sequence(converters.keys.CONVERT_TO_STRING)

        # Run
        converter = registry[key](None)
        value = converter("1")

        # Verify
        assert value == "1"

    def test_eq(self):
        assert converters.keys.Sequence(0) == converters.keys.Sequence(0)
        assert not (converters.keys.Sequence(1) == converters.keys.Sequence(0))
        assert not (converters.keys.Sequence(1) == 1)


class TestIdentity(object):
    _sentinel = object()

    @pytest.fixture(scope="class")
    def registry(self):
        return converters.ConverterFactoryRegistry(
            (converters.StandardConverter(),)
        )

    @pytest.fixture(scope="class")
    def key(self):
        return converters.keys.Identity()

    @pytest.mark.parametrize(
        "value, expected",
        [
            (1, 1),
            ("a", "a"),
            (_sentinel, _sentinel),
            ({"a": "b"}, {"a": "b"}),
            ([1, 2], [1, 2]),
        ],
    )
    def test_convert(self, registry, key, value, expected):
        converter = registry[key]()
        assert converter(value) == expected

    def test_eq(self):
        assert converters.keys.Identity() == converters.keys.Identity()


class TestRegistry(object):
    @pytest.mark.parametrize(
        "converter",
        # Try with both class and instance
        (converters.StandardConverter, converters.StandardConverter()),
    )
    def test_register_converter_factory_pass(self, converter):
        # Setup
        registry = register.Register()

        # Verify
        return_value = registry.register_converter_factory(converter)
        defaults = registry.get_converter_factories()
        assert return_value is converter
        assert len(defaults) == 1
        assert isinstance(defaults[0], converters.StandardConverter)

    def test_register_converter_factory_fail(self):
        # Setup
        registry = register.Register()

        # Verify failure when registered factory is not proper type.
        with pytest.raises(TypeError):
            registry.register_converter_factory(object())


class TestTypingConverter(object):
    singleton = converters.TypingConverter()
    inject_methods = pytest.mark.parametrize(
        "method, use_typing",
        [
            (singleton.create_request_body_converter, True),
            (singleton.create_request_body_converter, False),
            (singleton.create_response_body_converter, True),
            (singleton.create_response_body_converter, False),
        ],
    )

    @inject_methods
    def test_methods(self, method, use_typing):
        List, Dict = converters.typing_._get_types(use_typing)

        # Verify with sequence
        converter = method(List[str])
        assert isinstance(converter, converters.typing_.ListConverter)

        # Verify with mapping
        converter = method(Dict[str, str])
        assert isinstance(converter, converters.typing_.DictConverter)

        # Verify with unsupported value
        converter = method(None)
        assert converter is None

        # Verify with unsupported type
        if use_typing:
            converter = method(typing.Set[str])
            assert converter is None

    def test_list_converter(self):
        # Setup
        converter = converters.typing_.ListConverter(str)
        converter.set_chain(lambda x: x)

        # Verify
        output = converter([1, 2, 3])
        assert output == ["1", "2", "3"]

        # Verify with non-list: use element converter
        output = converter(1)
        assert output == ["1"]

    def test_dict_converter(self):
        # Setup
        converter = converters.typing_.DictConverter(int, str)
        converter.set_chain(lambda x: x)

        # Verify
        output = converter({"1": 1, "2": 2})
        assert output == {1: "1", 2: "2"}

        # Verify with non-map: use value converter
        output = converter(1)
        assert output == "1"


@pytest.mark.skipif(
    sys.version_info < (3, 6), reason="requires python3.6 or higher"
)
class TestPydanticConverter(object):
    @pytest.fixture
    def pydantic_model_mock(self, mocker):
        class Model(pydantic.BaseModel):
            def __new__(cls, *args, **kwargs):
                return model

        model = mocker.Mock(spec=Model)
        return model, Model

    def test_init_without_pydantic(self, mocker):
        mocker.patch.object(
            converters.PydanticConverter,
            "pydantic",
            new_callable=mocker.PropertyMock,
            return_value=None,
        )

        with pytest.raises(ImportError):
            converters.PydanticConverter()

    def test_create_request_body_converter(self, pydantic_model_mock):
        expected_result = {"id": 0}
        request_body = {}

        model_mock, model = pydantic_model_mock
        model_mock.dict.return_value = expected_result

        converter = converters.PydanticConverter()
        request_converter = converter.create_request_body_converter(model)

        result = request_converter.convert(request_body)

        assert result == expected_result
        model_mock.dict.assert_called_once()
        model_mock.dict.assert_called_once()

    def test_convert_complex_model(self):
        from json import loads
        from datetime import datetime

        class ComplexModel(pydantic.BaseModel):
            when = datetime.utcnow()  # type: datetime
            where = "http://example.com"  # type: pydantic.AnyUrl
            some = [1]  # type: typing.List[int]

        model = ComplexModel()
        request_body = {}
        expected_result = loads(model.json())

        converter = converters.PydanticConverter()
        request_converter = converter.create_request_body_converter(
            ComplexModel
        )

        result = request_converter.convert(request_body)

        assert result == expected_result

    def test_create_request_body_converter_with_original_model(
        self, pydantic_model_mock
    ):
        expected_result = {"id": 0}

        model_mock, model = pydantic_model_mock
        model_mock.dict.return_value = expected_result

        request_body = model()

        converter = converters.PydanticConverter()
        request_converter = converter.create_request_body_converter(model)

        result = request_converter.convert(request_body)

        assert result == expected_result
        model_mock.dict.assert_called_once()

    def test_create_request_body_converter_without_schema(self, mocker):
        expected_result = None
        converter = converters.PydanticConverter()

        result = converter.create_request_body_converter(mocker.sentinel)

        assert result is expected_result

    def test_create_response_body_converter(self, mocker, pydantic_model_mock):
        expected_result = "data"
        model_mock, model = pydantic_model_mock

        parse_obj_mock = mocker.patch.object(
            model, "parse_obj", return_value=expected_result
        )

        response = mocker.Mock(spec=["json"])
        response.json.return_value = {}

        converter = converters.PydanticConverter()
        c = converter.create_response_body_converter(model)

        result = c.convert(response)

        response.json.assert_called_once()
        parse_obj_mock.assert_called_once_with(response.json())
        assert result == expected_result

    def test_create_response_body_converter_invalid_response(
        self, mocker, pydantic_model_mock
    ):
        data = {"quick": "fox"}
        _, model = pydantic_model_mock

        parse_obj_mock = mocker.patch.object(
            model, "parse_obj", side_effect=pydantic.ValidationError([], model)
        )

        converter = converters.PydanticConverter()
        c = converter.create_response_body_converter(model)

        with pytest.raises(pydantic.ValidationError):
            c.convert(data)

        parse_obj_mock.assert_called_once_with(data)

    def test_create_response_body_converter_without_schema(self):
        expected_result = None
        converter = converters.PydanticConverter()

        result = converter.create_response_body_converter("not a schema")

        assert result is expected_result

    def test_create_string_converter(self, pydantic_model_mock):
        expected_result = None
        _, model = pydantic_model_mock
        converter = converters.PydanticConverter()

        c = converter.create_string_converter(model, None)

        assert c is expected_result

    @pytest.mark.parametrize(
        "pydantic_installed,expected",
        [
            pytest.param(
                True, [converters.PydanticConverter], id="pydantic_installed"
            ),
            pytest.param(None, [], id="pydantic_not_installed"),
        ],
    )
    def test_register(self, pydantic_installed, expected):
        converter = converters.PydanticConverter
        converter.pydantic = pydantic_installed

        register_ = []
        converter.register_if_necessary(register_.append)

        assert register_ == expected



# Standard library imports
import contextlib

# Third-party imports
import pytest

# Local imports
from uplink.clients import (
    AiohttpClient,
    interfaces,
    requests_,
    twisted_,
    register,
    io,
)

try:
    from uplink.clients import aiohttp_
except (ImportError, SyntaxError):
    aiohttp_ = None


requires_python34 = pytest.mark.skipif(
    not aiohttp_, reason="Requires Python 3.4 or above"
)


@contextlib.contextmanager
def _patch(obj, attr, value):
    if obj is not None:
        old_value = getattr(obj, attr)
        setattr(obj, attr, value)
    yield
    if obj is not None:
        setattr(obj, attr, old_value)


def test_get_default_client_with_non_callable():
    # Setup
    old_default = register.get_default_client()
    register.set_default_client("client")
    default_client = register.get_default_client()
    register.set_default_client(old_default)

    # Verify: an object that is not callable should be returned as set.
    assert default_client == "client"


def test_get_client_with_http_client_adapter_subclass():
    class HttpClientAdapterMock(interfaces.HttpClientAdapter):
        pass

    client = register.get_client(HttpClientAdapterMock)
    assert isinstance(client, HttpClientAdapterMock)


def test_get_client_with_unrecognized_key():
    assert register.get_client("no client for this key") is None


class TestRequests(object):
    def test_get_client(self, mocker):
        import requests

        session_mock = mocker.Mock(spec=requests.Session)
        client = register.get_client(session_mock)
        assert isinstance(client, requests_.RequestsClient)

    def test_init_with_kwargs(self):
        session = requests_.RequestsClient._create_session(
            cred=("username", "password")
        )
        assert session.cred == ("username", "password")

    def test_client_send(self, mocker):
        # Setup
        import requests

        session_mock = mocker.Mock(spec=requests.Session)
        session_mock.request.return_value = "response"
        callback = mocker.stub()
        client = requests_.RequestsClient(session_mock)

        # Run
        response = client.send(("method", "url", {}))

        # Verify
        session_mock.request.assert_called_with(method="method", url="url")

        # Run callback
        client.apply_callback(callback, response)
        callback.assert_called_with(session_mock.request.return_value)

    def test_dont_close_provided_session(self, mocker):
        # Setup
        import requests
        import gc

        session_mock = mocker.Mock(spec=requests.Session)
        session_mock.request.return_value = "response"

        # Run
        client = requests_.RequestsClient(session_mock)
        client.send(("method", "url", {}))
        del client
        gc.collect()

        assert session_mock.close.call_count == 0

    def test_close_auto_generated_session(self, mocker):
        # Setup
        import requests
        import gc

        session_mock = mocker.Mock(spec=requests.Session)
        session_mock.request.return_value = "response"
        session_cls_mock = mocker.patch("requests.Session")
        session_cls_mock.return_value = session_mock

        # Run
        client = requests_.RequestsClient()
        client.send(("method", "url", {}))
        del client
        gc.collect()

        assert session_mock.close.call_count == 1

    def test_exceptions(self):
        import requests

        exceptions = requests_.RequestsClient.exceptions

        with pytest.raises(exceptions.BaseClientException):
            raise requests.RequestException()

        with pytest.raises(exceptions.BaseClientException):
            # Test polymorphism
            raise requests.exceptions.InvalidURL()

        with pytest.raises(exceptions.ConnectionError):
            raise requests.exceptions.ConnectionError()

        with pytest.raises(exceptions.ConnectionTimeout):
            raise requests.exceptions.ConnectTimeout()

        with pytest.raises(exceptions.ServerTimeout):
            raise requests.exceptions.ReadTimeout()

        with pytest.raises(exceptions.SSLError):
            raise requests.exceptions.SSLError()

        with pytest.raises(exceptions.InvalidURL):
            raise requests.exceptions.InvalidURL()

    def test_io(self):
        assert isinstance(requests_.RequestsClient.io(), io.BlockingStrategy)


class TestTwisted(object):
    def test_init_without_client(self):
        twisted = twisted_.TwistedClient()
        assert isinstance(twisted._proxy, requests_.RequestsClient)

    def test_create_requests_no_twisted(self, http_client_mock):
        with _patch(twisted_, "threads", None):
            with pytest.raises(NotImplementedError):
                twisted_.TwistedClient(http_client_mock)

    def test_client_send(self, mocker, http_client_mock):
        deferToThread = mocker.patch.object(twisted_.threads, "deferToThread")
        request = twisted_.TwistedClient(http_client_mock)
        request.send((1, 2, 3))
        deferToThread.assert_called_with(http_client_mock.send, (1, 2, 3))

    def test_client_callback(self, mocker, http_client_mock):
        # Setup
        callback = mocker.stub()
        deferred = mocker.Mock()
        deferToThread = mocker.patch.object(twisted_.threads, "deferToThread")
        deferToThread.return_value = deferred
        client = twisted_.TwistedClient(http_client_mock)

        # Run
        response = client.apply_callback(callback, 1)

        # Verify
        assert response is deferred
        deferToThread.assert_called_with(
            http_client_mock.apply_callback, callback, 1
        )

    def test_exceptions(self, http_client_mock):
        twisted_client = twisted_.TwistedClient(http_client_mock)
        assert http_client_mock.exceptions == twisted_client.exceptions

    def test_io(self):
        assert isinstance(twisted_.TwistedClient.io(), io.TwistedStrategy)


@pytest.fixture
def aiohttp_session_mock(mocker):
    import aiohttp

    return mocker.Mock(spec=aiohttp.ClientSession)


class TestAiohttp(object):
    @requires_python34
    def test_init_when_aiohttp_is_not_installed(self):
        with _patch(aiohttp_, "aiohttp", None):
            with pytest.raises(NotImplementedError):
                AiohttpClient()

    @requires_python34
    def test_init_with_session_None(self, mocker):
        mocker.spy(AiohttpClient, "_create_session")
        AiohttpClient(kwarg1="value")
        AiohttpClient._create_session.assert_called_with(kwarg1="value")

    @requires_python34
    def test_get_client(self, aiohttp_session_mock):
        client = register.get_client(aiohttp_session_mock)
        assert isinstance(client, aiohttp_.AiohttpClient)

    @requires_python34
    def test_request_send(self, mocker, aiohttp_session_mock):
        # Setup
        import asyncio

        expected_response = mocker.Mock()

        @asyncio.coroutine
        def request(*args, **kwargs):
            return expected_response

        aiohttp_session_mock.request = request
        client = aiohttp_.AiohttpClient(aiohttp_session_mock)

        # Run
        response = client.send((1, 2, {}))
        loop = asyncio.get_event_loop()
        value = loop.run_until_complete(asyncio.ensure_future(response))

        # Verify
        assert value == expected_response

    @requires_python34
    def test_callback(self, mocker, aiohttp_session_mock):
        # Setup
        import asyncio

        expected_response = mocker.Mock(spec=aiohttp_.aiohttp.ClientResponse)

        @asyncio.coroutine
        def request(*args, **kwargs):
            return expected_response

        aiohttp_session_mock.request = request
        client = aiohttp_.AiohttpClient(aiohttp_session_mock)
        client._sync_callback_adapter = asyncio.coroutine

        # Run
        globals_ = dict(globals(), client=client)
        locals_ = {"asyncio": asyncio}
        exec(
            """@asyncio.coroutine
def call():
    response = yield from client.send((1, 2, {}))
    response = yield from client.apply_callback(lambda x: 2, response)
    return response
""",
            globals_,
            locals_,
        )

        loop = asyncio.get_event_loop()
        value = loop.run_until_complete(
            asyncio.ensure_future(locals_["call"]())
        )

        # Verify
        assert value == 2

    @requires_python34
    def test_wrap_callback(self, mocker):
        import asyncio

        # Setup
        c = AiohttpClient()
        mocker.spy(c, "_sync_callback_adapter")

        # Run: with callback that is not a coroutine
        def callback(*_):
            pass

        c.wrap_callback(callback)

        # Verify: Should wrap it
        c._sync_callback_adapter.assert_called_with(callback)

        # Run: with coroutine callback
        coroutine_callback = asyncio.coroutine(callback)
        assert c.wrap_callback(coroutine_callback) is coroutine_callback

    @requires_python34
    def test_threaded_callback(self, mocker):
        import asyncio

        def callback(response):
            return response

        # Mock response.
        response = mocker.Mock(spec=aiohttp_.aiohttp.ClientResponse)
        response.text = asyncio.coroutine(mocker.stub())

        # Run
        new_callback = aiohttp_.threaded_callback(callback)
        return_value = new_callback(response)
        loop = asyncio.get_event_loop()
        value = loop.run_until_complete(asyncio.ensure_future(return_value))

        # Verify
        response.text.assert_called_with()
        assert value == response

        # Run: Verify with callback that returns new value
        def callback(*_):
            return 1

        new_callback = aiohttp_.threaded_callback(callback)
        awaitable = new_callback(response)
        loop = asyncio.get_event_loop()
        value = loop.run_until_complete(asyncio.ensure_future(awaitable))
        assert value == 1
        assert response.text.called

        # Run: Verify with response that is not ClientResponse (should not be wrapped)
        response = mocker.Mock()
        awaitable = new_callback(response)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(asyncio.ensure_future(awaitable))
        assert not response.text.called

    @requires_python34
    def test_threaded_coroutine(self):
        # Setup
        import asyncio

        @asyncio.coroutine
        def coroutine():
            return 1

        threaded_coroutine = aiohttp_.ThreadedCoroutine(coroutine)

        # Run -- should block
        response = threaded_coroutine()

        # Verify
        assert response == 1

    @requires_python34
    def test_threaded_response(self, mocker):
        # Setup
        import asyncio

        @asyncio.coroutine
        def coroutine():
            return 1

        def not_a_coroutine():
            return 2

        response = mocker.Mock()
        response.coroutine = coroutine
        response.not_coroutine = not_a_coroutine
        threaded_response = aiohttp_.ThreadedResponse(response)

        # Run
        threaded_coroutine = threaded_response.coroutine
        return_value = threaded_coroutine()

        # Verify
        assert isinstance(threaded_coroutine, aiohttp_.ThreadedCoroutine)
        assert return_value == 1
        assert threaded_response.not_coroutine is not_a_coroutine

    @requires_python34
    def test_create(self, mocker):
        # Setup
        import asyncio

        session_cls_mock = mocker.patch("aiohttp.ClientSession")
        positionals = [1]
        keywords = {"keyword": 2}

        # Run: Create client
        client = aiohttp_.AiohttpClient.create(*positionals, **keywords)

        # Verify: session hasn't been created yet.
        assert not session_cls_mock.called

        # Run: Get session
        loop = asyncio.get_event_loop()
        loop.run_until_complete(asyncio.ensure_future(client.session()))

        # Verify: session created with args
        session_cls_mock.assert_called_with(*positionals, **keywords)

    @requires_python34
    def test_close_auto_created_session(self, mocker):
        # Setup
        import asyncio
        import gc
        import aiohttp

        mock_session = mocker.Mock(spec=aiohttp.ClientSession)
        session_cls_mock = mocker.patch("aiohttp.ClientSession")
        session_cls_mock.return_value = mock_session

        positionals = [1]
        keywords = {"keyword": 2}

        # Run: Create client
        client = aiohttp_.AiohttpClient.create(*positionals, **keywords)

        # Run: Get session
        loop = asyncio.get_event_loop()
        loop.run_until_complete(asyncio.ensure_future(client.session()))

        # Verify: session created with args
        session_cls_mock.assert_called_with(*positionals, **keywords)
        del client
        gc.collect()
        session_cls_mock.return_value.close.assert_called_with()

    @requires_python34
    def test_exceptions(self):
        import aiohttp

        exceptions = aiohttp_.AiohttpClient.exceptions

        with pytest.raises(exceptions.BaseClientException):
            raise aiohttp.ClientError()

        with pytest.raises(exceptions.BaseClientException):
            # Test polymorphism
            raise aiohttp.InvalidURL("invalid")

        with pytest.raises(exceptions.ConnectionError):
            raise aiohttp.ClientConnectionError()

        with pytest.raises(exceptions.ConnectionTimeout):
            raise aiohttp.ClientConnectorError.__new__(
                aiohttp.ClientConnectorError
            )

        with pytest.raises(exceptions.ServerTimeout):
            raise aiohttp.ServerTimeoutError()

        with pytest.raises(exceptions.SSLError):
            raise aiohttp.ClientSSLError.__new__(aiohttp.ClientSSLError)

        with pytest.raises(exceptions.InvalidURL):
            raise aiohttp.InvalidURL("invalid")

    @requires_python34
    def test_io(self):
        assert isinstance(aiohttp_.AiohttpClient.io(), io.AsyncioStrategy)



# Third-party imports
import pytest

# Local imports
import uplink

# Constants
BASE_URL = "https://api.github.com"


class GitHubError(Exception):
    pass


@uplink.response_handler
def github_error(response):
    if "errors" in response.json():
        raise GitHubError()
    return response


@uplink.timeout(10)
class GitHubService(uplink.Consumer):
    @github_error
    @uplink.json
    @uplink.post("graphql", args=(uplink.Body,))
    def graphql(self, **body):
        pass

    @uplink.returns.json(member=("data", "repository"))
    @uplink.args(body=uplink.Body)
    @graphql
    def get_repository(self, **body):
        pass

    @uplink.returns.json(member=("data", "repository"))
    @graphql.extend("graphql2", args=(uplink.Body,))
    def get_repository2(self, **body):
        pass


def test_get_repository(mock_client, mock_response):
    data = {
        "query": """\
query {
    repository(owner: "prkumar", name: "uplink") {
        nameWithOwner
    }
}"""
    }
    result = {"data": {"repository": {"nameWithOwner": "prkumar/uplink"}}}
    mock_response.with_json(result)
    mock_client.with_response(mock_response)
    github = GitHubService(base_url=BASE_URL, client=mock_client)
    response = github.get_repository(**data)
    request = mock_client.history[0]
    assert request.method == "POST"
    assert request.base_url == BASE_URL
    assert request.endpoint == "/graphql"
    assert request.timeout == 10
    assert request.json == data
    assert response == result["data"]["repository"]


def test_get_repository2_failure(mock_client, mock_response):
    data = {
        "query": """\
query {
    repository(owner: "prkumar", name: "uplink") {
        nameWithOwner
    }
}"""
    }
    result = {
        "data": {"repository": None},
        "errors": [
            {
                "type": "NOT_FOUND",
                "path": ["repository"],
                "locations": [{"line": 7, "column": 3}],
                "message": "Could not resolve to a User with the username 'prkussmar'.",
            }
        ],
    }
    mock_response.with_json(result)
    mock_client.with_response(mock_response)
    github = GitHubService(base_url=BASE_URL, client=mock_client)
    with pytest.raises(GitHubError):
        github.get_repository2(**data)
    request = mock_client.history[0]
    assert request.method == "POST"
    assert request.base_url == BASE_URL
    assert request.endpoint == "/graphql2"
    assert request.timeout == 10



# Standard library imports
from collections import abc
import functools

# Local imports
from uplink import (
    arguments,
    converters,
    decorators,
    exceptions,
    interfaces,
    returns,
    utils,
)

__all__ = ["get", "head", "put", "post", "patch", "delete"]


class MissingArgumentAnnotations(exceptions.InvalidRequestDefinition):
    message = "Missing annotation for argument(s): '%s'."
    implicit_message = " (Implicit path variables: '%s')"

    def __init__(self, missing, path_variables):
        missing, path_variables = list(missing), list(path_variables)
        self.message = self.message % "', '".join(missing)
        if path_variables:  # pragma: no cover
            self.message += self.implicit_message % "', '".join(path_variables)


class MissingUriVariables(exceptions.InvalidRequestDefinition):
    message = "On uri template %s, some variables are not handled: %s"

    def __init__(self, uri, remaining_variables):
        self.message = self.message % (uri, "', '".join(remaining_variables))


class URIDefinitionBuilder(interfaces.UriDefinitionBuilder):
    def __init__(self, uri):
        self._uri = uri
        self._is_dynamic = False
        self._uri_variables = set()

    @property
    def template(self):
        return self._uri

    @property
    def is_static(self):
        return self._uri is not None

    @property
    def is_dynamic(self):
        return self._is_dynamic

    @is_dynamic.setter
    def is_dynamic(self, is_dynamic):
        if self.is_static:
            raise ValueError(
                "Failed to set dynamic URI as URI is already defined: %s"
                % self._uri
            )
        self._is_dynamic = is_dynamic

    def add_variable(self, name):
        if self.is_static and name not in self.remaining_variables:
            raise ValueError(
                "`%s` is not a variable on the uri %s" % (name, self._uri)
            )
        self._uri_variables.add(name)

    @property
    def remaining_variables(self):
        return utils.URIBuilder.variables(self._uri) - self._uri_variables

    def build(self):
        if self.remaining_variables:
            raise MissingUriVariables(self._uri, self.remaining_variables)
        return self._uri


class RequestDefinitionBuilder(interfaces.RequestDefinitionBuilder):
    def __init__(
        self,
        func,
        method,
        uri,
        argument_handler_builder,
        method_handler_builder,
    ):
        self._func = func
        self._method = method
        self._uri = uri
        self._argument_handler_builder = argument_handler_builder
        self._method_handler_builder = method_handler_builder
        self._return_type = None

        self._argument_handler_builder.listener = self._notify
        self._method_handler_builder.listener = self._notify

    def _notify(self, annotation):
        annotation.modify_request_definition(self)

    @property
    def method(self):
        return self._method

    @property
    def uri(self):
        return self._uri

    @property
    def argument_handler_builder(self):
        return self._argument_handler_builder

    @property
    def method_handler_builder(self):
        return self._method_handler_builder

    @property
    def return_type(self):
        return self._return_type

    @return_type.setter
    def return_type(self, return_type):
        self._return_type = return_type

    def __call__(self, uri=None, args=()):
        """
        Applies the decorators, HTTP method, and optionally the URI
        of this consumer method to the decorated method.

        This makes the request definition reusable and can help
        minimize duplication across similar consumer methods.

        Examples:

            Define request templates:

            .. code-block:: python

                from uplink import Consumer, get, json, returns

                @returns.json
                @json
                @get
                def get_json():
                    \"""GET request that consumes and produces JSON.\"""

                class GitHub(Consumer):
                    @get_json("/users/{user}")
                    def get_user(self, user):
                         \"""Fetches a specific GitHub user.\"""

            Remove duplication across definitions of similar consumer
            methods, whether or not the methods are defined in the same
            class:

            .. code-block:: python

                from uplink import Consumer, get, params, timeout

                class GitHub(Consumer):
                    @timeout(10)
                    @get("/users/{user}/repos")
                    def get_user_repos(self, user):
                        \"""Retrieves the repos that the user owns.\"""

                    # Extends the above method to define a variant:
                    @params(type="member")
                    @get_user_repos
                    def get_repos_for_collaborator(self, user):
                        \"""
                        Retrieves the repos for which the given user is
                        a collaborator.
                        \"""

                class EnhancedGitHub(Github):
                    # Updates the return type of an inherited method.
                    @GitHub.get_user_repos
                    def get_user_repos(self, user) -> List[Repo]:
                        \"""Retrieves the repos that the user owns.\"""

        Args:
            uri (str, optional): the request's relative path
            args: a list or mapping of function annotations (e.g.
                  :class:`uplink.Path`) corresponding to the decorated
                  function's arguments
        """
        return self.extend(uri, args)

    def extend(self, uri=None, args=()):
        factory = HttpMethodFactory(
            method=self.method, request_definition_builder_factory=self._extend
        )

        if callable(uri):
            return factory(self.uri.template, args)(uri)
        else:
            uri = self.uri.template if uri is None else uri
            return factory(uri, args)

    def _extend(self, func, method, uri, arg_handler, _):
        builder = RequestDefinitionBuilder(
            # Extended consumer methods should only inherit the decorators and
            # not any function annotations, since the new method can have a
            # different signature than the original.
            func,
            method,
            uri,
            arg_handler,
            self.method_handler_builder.copy(),
        )
        builder.return_type = self.return_type
        return builder

    def copy(self):
        builder = RequestDefinitionBuilder(
            self._func,
            self._method,
            self._uri,
            self._argument_handler_builder.copy(),
            self._method_handler_builder.copy(),
        )
        builder.return_type = self.return_type
        return builder

    def _auto_fill_remaining_arguments(self):
        uri_vars = set(self.uri.remaining_variables)
        missing = list(self.argument_handler_builder.missing_arguments)
        still_missing = set(missing) - uri_vars

        # Preserve order of function parameters.
        matching = [p for p in missing if p in uri_vars]

        if still_missing:
            raise MissingArgumentAnnotations(still_missing, matching)

        path_vars = dict.fromkeys(matching, arguments.Path)
        self.argument_handler_builder.set_annotations(path_vars)

    def update_wrapper(self, wrapper):
        functools.update_wrapper(wrapper, self._func)

    def build(self):
        if not self._argument_handler_builder.is_done():
            self._auto_fill_remaining_arguments()
        argument_handler = self._argument_handler_builder.build()
        method_handler = self._method_handler_builder.build()
        uri = self._uri.build()
        return RequestDefinition(
            self._method,
            uri,
            self._return_type,
            argument_handler,
            method_handler,
        )


class RequestDefinition(interfaces.RequestDefinition):
    def __init__(
        self, method, uri, return_type, argument_handler, method_handler
    ):
        self._method = method
        self._uri = uri
        self._return_type = return_type
        self._argument_handler = argument_handler
        self._method_handler = method_handler

    @property
    def argument_annotations(self):
        return tuple(self._argument_handler.annotations)

    @property
    def method_annotations(self):
        return tuple(self._method_handler.annotations)

    def make_converter_registry(self, converters_):
        return converters.ConverterFactoryRegistry(converters_, self)

    def define_request(self, request_builder, func_args, func_kwargs):
        request_builder.method = self._method
        request_builder.relative_url = self._uri
        request_builder.return_type = self._return_type
        self._argument_handler.handle_call(
            request_builder, func_args, func_kwargs
        )
        self._method_handler.handle_builder(request_builder)


class HttpMethodFactory(object):
    def __init__(
        self,
        method,
        request_definition_builder_factory=RequestDefinitionBuilder,
    ):
        self._method = method
        self._request_definition_builder_factory = (
            request_definition_builder_factory
        )

    def __call__(self, uri=None, args=()):
        if callable(uri) and not args:
            return HttpMethod(self._method)(
                uri, self._request_definition_builder_factory
            )
        else:
            return functools.partial(
                HttpMethod(self._method, uri, args),
                request_definition_builder_factory=self._request_definition_builder_factory,
            )


class HttpMethod(object):
    @staticmethod
    def _add_args(obj):
        return obj

    def __init__(self, method, uri=None, args=None):
        self._method = method
        self._uri = uri

        # Register argument annotations
        if args:
            is_map = isinstance(args, abc.Mapping)
            args, kwargs = ((), args) if is_map else (args, {})
            self._add_args = decorators.args(*args, **kwargs)

    def __call__(
        self, func, request_definition_builder_factory=RequestDefinitionBuilder
    ):
        spec = utils.get_arg_spec(func)
        arg_handler = arguments.ArgumentAnnotationHandlerBuilder(
            func, spec.args
        )
        builder = request_definition_builder_factory(
            func,
            self._method,
            URIDefinitionBuilder(self._uri),
            arg_handler,
            decorators.MethodAnnotationHandlerBuilder(),
        )

        # Need to add the annotations after constructing the request
        # definition builder so it has a chance to attach its listener.
        arg_handler.set_annotations(spec.annotations)

        # Use return value type hint as expected return type
        if spec.return_annotation is not None:
            builder = returns.schema(spec.return_annotation)(builder)
        builder = self._add_args(builder)
        return builder


get = HttpMethodFactory("GET").__call__
head = HttpMethodFactory("HEAD").__call__
put = HttpMethodFactory("PUT").__call__
post = HttpMethodFactory("POST").__call__
patch = HttpMethodFactory("PATCH").__call__
delete = HttpMethodFactory("DELETE").__call__



# Local imports
from uplink import session


def test_base_url(uplink_builder_mock):
    # Setup
    uplink_builder_mock.base_url = "https://api.github.com"
    sess = session.Session(uplink_builder_mock)

    # Run & Verify
    assert uplink_builder_mock.base_url == sess.base_url


def test_headers(uplink_builder_mock):
    # Setup
    sess = session.Session(uplink_builder_mock)

    # Run
    sess.headers["key"] = "value"

    # Verify
    assert uplink_builder_mock.add_hook.called
    assert sess.headers == {"key": "value"}


def test_params(uplink_builder_mock):
    # Setup
    sess = session.Session(uplink_builder_mock)

    # Run
    sess.params["key"] = "value"

    # Verify
    assert uplink_builder_mock.add_hook.called
    assert sess.params == {"key": "value"}


def test_auth(uplink_builder_mock):
    # Setup
    uplink_builder_mock.auth = ("username", "password")
    sess = session.Session(uplink_builder_mock)

    # Run & Verify
    assert uplink_builder_mock.auth == sess.auth


def test_auth_set(uplink_builder_mock):
    # Setup
    sess = session.Session(uplink_builder_mock)

    # Run
    sess.auth = ("username", "password")

    # Verify
    assert ("username", "password") == uplink_builder_mock.auth


def test_context(uplink_builder_mock):
    # Setup
    sess = session.Session(uplink_builder_mock)

    # Run
    sess.context["key"] = "value"

    # Verify
    assert uplink_builder_mock.add_hook.called
    assert sess.context == {"key": "value"}



# Standard library imports
import collections

# Third-party imports
import pytest

# Local imports
from uplink import clients, converters, hooks, interfaces, helpers
from uplink.clients.exceptions import Exceptions


@pytest.fixture
def http_client_mock(mocker):
    client = mocker.Mock(spec=clients.interfaces.HttpClientAdapter)
    return client


@pytest.fixture
def transaction_hook_mock(mocker):
    return mocker.Mock(spec=hooks.TransactionHook)


@pytest.fixture
def converter_mock(mocker):
    return mocker.Mock(spec=converters.interfaces.Converter)


@pytest.fixture
def converter_factory_mock(mocker):
    return mocker.Mock(spec=converters.interfaces.Factory)


@pytest.fixture
def annotation_mock(mocker):
    return mocker.Mock(spec=interfaces.Annotation)


@pytest.fixture
def annotation_handler_builder_mock(mocker):
    return mocker.Mock(spec=interfaces.AnnotationHandlerBuilder)


@pytest.fixture
def annotation_handler_mock(mocker):
    return mocker.Mock(spec=interfaces.AnnotationHandler)


@pytest.fixture
def request_definition_builder(mocker):
    return mocker.Mock(spec=interfaces.RequestDefinitionBuilder)


@pytest.fixture
def request_definition(mocker):
    return mocker.Mock(spec=interfaces.RequestDefinition)


@pytest.fixture
def uplink_builder_mock(mocker):
    return mocker.Mock(spec=interfaces.CallBuilder)


@pytest.fixture
def request_builder(mocker):
    builder = mocker.MagicMock(spec=helpers.RequestBuilder)
    builder.info = collections.defaultdict(dict)
    builder.context = {}
    builder.get_converter.return_value = lambda x: x
    builder.client.exceptions = Exceptions()
    return builder



# Third-party imports
import pytest

# Local imports
from uplink import _extras


def test_load_entry_points(mocker):
    # Setup
    func = mocker.stub()
    iter_entry_points = mocker.stub()
    entry_point = mocker.Mock()
    entry_point.name = "plugin-name"
    entry_point.load.return_value = "plugin-value"
    iter_entry_points.return_value = [entry_point]
    entry_points = {"entry-point": func}

    # Run
    _extras.load_entry_points(
        _entry_points=entry_points, _iter_entry_points=iter_entry_points
    )

    # Verify
    func.assert_called_with("plugin-value")


def test_install(mocker):
    # Setup
    func = mocker.stub()
    installers = {list: func}
    obj = []

    # Run: Success
    _extras.install(obj, _installers=installers)

    # Verify
    func.assert_called_with(obj)

    # Run & Verify Failure
    with pytest.raises(TypeError):
        _extras.install(dict(), _installers=installers)



# Standard library imports
import sys

# Local imports
from uplink import utils

is_py2 = sys.version_info[0] == 2


def test_get_arg_spec():
    if is_py2:
        code = "def func(pos1, *args, **kwargs): pass"
    else:
        code = "def func(pos1, *args: 2, **kwargs: 3) -> 4: pass"
    exec(code, globals(), locals())
    signature = utils.get_arg_spec(locals()["func"])
    assert isinstance(signature, utils.Signature)
    assert signature.args == ["pos1", "args", "kwargs"]
    if not is_py2:
        assert signature.annotations == {"args": 2, "kwargs": 3}
        assert signature.return_annotation == 4


def test_call_args():
    def func(pos1, *args, **kwargs):
        pass

    call_args = utils.get_call_args(func, 1, 2, named=3)
    assert call_args == {"pos1": 1, "args": (2,), "kwargs": {"named": 3}}


class TestURIBuilder(object):
    def test_variables_not_string(self):
        assert utils.URIBuilder.variables(None) == set()

    def test_set_variable(self):
        builder = utils.URIBuilder("/path/to/{variable}")
        assert builder.build() == "/path/to/"
        builder.set_variable(variable="resource")
        assert builder.build() == "/path/to/resource"

    def test_remaining_variables(self):
        builder = utils.URIBuilder("{variable}")
        assert builder.remaining_variables() == set(["variable"])
        builder.set_variable(variable="resource")
        assert len(builder.remaining_variables()) == 0



# Standard library imports
from collections import abc

# Local imports
from uplink import compat


class IllegalRequestStateTransition(RuntimeError):
    """An improper request state transition was attempted."""

    def __init__(self, state, transition):
        self._state = state
        self._transition = transition

    def __str__(self):
        return (
            "Illegal transition [%s] from request state [%s]: this is "
            "possibly due to a badly designed RequestTemplate."
            % (self._transition, self._state)
        )


class InvokeCallback(object):
    """
    Callbacks to continue the running request execution after invoking
    a function using the underlying I/O model.
    """

    def on_success(self, result):
        """
        Handles a successful invocation.

        Args:
            result: The invocation's return value.
        """
        raise NotImplementedError

    def on_failure(self, exc_type, exc_val, exc_tb):
        """
        Handles a failed invocation.

        Args:
            exc_type: The exception class.
            exc_val: The exception object.
            exc_tb: The exception's stacktrace.
        """
        raise NotImplementedError


class SleepCallback(object):
    """
    Callbacks to continue the running request execution after an
    intended pause.
    """

    def on_success(self):
        """Handles a successful pause."""
        raise NotImplementedError

    def on_failure(self, exc_type, exc_val, exc_tb):
        """
        Handles a failed pause.

        Args:
            exc_type: The exception class.
            exc_val: The exception object.
            exc_tb: The exception's stacktrace.
        """
        raise NotImplementedError


class Executable(abc.Iterator):
    """An abstraction for iterating over the execution of a request."""

    def __next__(self):
        return self.execute()

    next = __next__

    def execute(self):
        """Continues the request's execution."""
        raise NotImplementedError


class RequestExecution(Executable):
    """A state machine representing the execution lifecycle of a request."""

    @property
    def state(self):
        """The current state of the request."""
        raise NotImplementedError

    def send(self, request, callback):
        """
        Sends the given request.

        Args:
            request: The intended request data to be sent.
            callback (InvokeCallback): A callback that resumes execution
                after the request is sent.
        """
        raise NotImplementedError

    def sleep(self, duration, callback):
        """
        Pauses the execution for the allotted duration.

        Args:
            duration: The number of seconds to delay execution.
            callback (:obj:`SleepCallback`): A callback that resumes
                execution after the delay.
        """
        raise NotImplementedError

    def finish(self, response):
        """
        Completes the execution.

        Args:
            response: The object to return to the execution's invoker.
        """
        raise NotImplementedError

    def fail(self, exc_type, exc_val, exc_tb):
        """
        Fails the execution with a specific error.

        Args:
            exc_type: The exception class.
            exc_val: The exception object.
            exc_tb: The exception's stacktrace.
        """
        raise NotImplementedError

    def execute(self):
        """Performs the next sequence of steps in the execution."""
        raise NotImplementedError

    def before_request(self, request):
        """Handles transitioning the execution before the request is sent."""
        raise NotImplementedError

    def after_response(self, request, response):
        """Handles transitioning the execution after a successful request."""
        raise NotImplementedError

    def after_exception(self, request, exc_type, exc_val, exc_tb):
        """Handles transitioning the execution after a failed request."""
        raise NotImplementedError

    def start(self, request):
        """Starts the request's execution."""
        raise NotImplementedError


class RequestState(object):
    @property
    def request(self):
        raise NotImplementedError

    def send(self, request):
        raise IllegalRequestStateTransition(self, "send")

    def prepare(self, request):
        raise IllegalRequestStateTransition(self, "prepare")

    def sleep(self, duration):
        raise IllegalRequestStateTransition(self, "sleep")

    def finish(self, response):
        raise IllegalRequestStateTransition(self, "finish")

    def fail(self, exc_type, exc_val, exc_tb):
        raise IllegalRequestStateTransition(self, "fail")

    def execute(self, execution):
        raise NotImplementedError


class RequestTemplate(object):
    """
    Hooks for managing the lifecycle of a request.

    To modify behavior of a specific part of the request, override the
    appropriate hook and return the intended transition from
    :mod:`uplink.clients.io.transitions`.

    To fallback to the default behavior, either don't override the hook
    or return :obj:`None` instead, in case of conditional overrides
    (e.g., retry the request if it has failed less than a certain number
    of times).
    """

    def before_request(self, request):
        """
        Handles the request before it is sent.

        Args:
            request: The prospective request data.

        Returns:
            ``None`` or a transition from
            :mod:`uplink.clients.io.transitions`.
        """

    def after_response(self, request, response):
        """
        Handles the response after a successful request.

        Args:
            request: The data sent to the server.
            response: The response returned by server.

        Returns:
            ``None`` or a transition from
            :mod:`uplink.clients.io.transitions`.
        """

    def after_exception(self, request, exc_type, exc_val, exc_tb):
        """
        Handles the error after a failed request.

        Args:
            request: The attempted request.
            exc_type: The exception class.
            exc_val: The exception object.
            exc_tb: The exception's stacktrace.

        Returns:
            ``None`` or a transition from
            :mod:`uplink.clients.io.transitions`.
        """


class Client(object):
    """An HTTP Client implementation."""

    def send(self, request):
        """
        Sends the given request.

        Args:
            request: The intended request data to be sent.
        """
        raise NotImplementedError

    def apply_callback(self, callback, response):
        """
        Invokes callback on the response.

        Args:
            callback (callable): a function that handles the response.
            response: data returned from a server after request.
        """
        raise NotImplementedError


class IOStrategy(object):
    """An adapter for a specific I/O model."""

    def invoke(self, func, args, kwargs, callback):
        """
        Invokes the given function using the underlying I/O model.

        Args:
            func (callback): The function to invoke.
            args: The function's positional arguments.
            kwargs: The function's keyword arguments.
            callback (:obj:`InvokeCallback`): A callback that resumes
                execution after the invocation completes.
        """
        raise NotImplementedError

    def sleep(self, duration, callback):
        """
        Pauses the execution for the allotted duration.

        Args:
            duration: The number of seconds to delay execution.
            callback (:obj:`SleepCallback`): A callback that resumes
                execution after the delay.
        """
        raise NotImplementedError

    def finish(self, response):
        """
        Completes the execution.

        Args:
            response: The object to return to the execution's invoker.
        """
        raise NotImplementedError

    def fail(self, exc_type, exc_val, exc_tb):
        """
        Fails the execution with a specific error.

        Args:
            exc_type: The exception class.
            exc_val: The exception object.
            exc_tb: The exception's stacktrace.
        """
        compat.reraise(exc_type, exc_val, exc_tb)

    def execute(self, executable):
        """
        Runs a request's execution to completion using the I/O framework
        of this strategy.
        """
        raise NotImplementedError



# Third-party imports
import pytest

# Local imports
from uplink import auth, builder, converters, exceptions, helpers
from uplink.clients import io


@pytest.fixture
def fake_service_cls(request_definition_builder, request_definition):
    class Service(object):
        builder = request_definition_builder

    request_definition_builder.build.return_value = request_definition
    return Service


@pytest.fixture
def uplink_builder(http_client_mock):
    b = builder.Builder()
    b.client = http_client_mock
    return b


class TestRequestPreparer(object):
    def test_prepare_request(self, mocker, request_builder):
        request_builder.method = "METHOD"
        request_builder.relative_url = "/example/path"
        request_builder.return_type = None
        request_builder.transaction_hooks = ()
        request_builder.request_template = "request_template"
        uplink_builder = mocker.Mock(spec=builder.Builder)
        uplink_builder.converters = ()
        uplink_builder.hooks = ()
        request_preparer = builder.RequestPreparer(uplink_builder)
        execution_builder = mocker.Mock(spec=io.RequestExecutionBuilder)
        request_preparer.prepare_request(request_builder, execution_builder)

        # Verify
        execution_builder.with_client.assert_called_with(uplink_builder.client)
        execution_builder.with_io.assert_called_with(uplink_builder.client.io())
        execution_builder.with_template(request_builder.request_template)

    def test_prepare_request_with_transaction_hook(
        self, mocker, uplink_builder, request_builder, transaction_hook_mock
    ):
        request_builder.method = "METHOD"
        request_builder.relative_url = "/example/path"
        request_builder.request_template = "request_template"
        uplink_builder.base_url = "https://example.com"
        request_builder.transaction_hooks = [transaction_hook_mock]
        request_preparer = builder.RequestPreparer(uplink_builder)
        execution_builder = mocker.Mock(spec=io.RequestExecutionBuilder)
        request_preparer.prepare_request(request_builder, execution_builder)

        # Verify
        transaction_hook_mock.audit_request.assert_called_with(
            None, request_builder
        )
        execution_builder.with_client.assert_called_with(uplink_builder.client)
        execution_builder.with_io.assert_called_with(uplink_builder.client.io())
        execution_builder.with_template(request_builder.request_template)

    def test_create_request_builder(self, mocker, request_definition):
        uplink_builder = mocker.Mock(spec=builder.Builder)
        uplink_builder.converters = ()
        uplink_builder.hooks = ()
        request_definition.make_converter_registry.return_value = {}
        request_preparer = builder.RequestPreparer(uplink_builder)
        request = request_preparer.create_request_builder(request_definition)
        assert isinstance(request, helpers.RequestBuilder)

    def test_create_request_builder_with_session_hooks(
        self, mocker, request_definition, transaction_hook_mock
    ):
        uplink_builder = mocker.Mock(spec=builder.Builder)
        uplink_builder.converters = ()
        uplink_builder.hooks = (transaction_hook_mock,)
        request_definition.make_converter_registry.return_value = {}
        request_preparer = builder.RequestPreparer(uplink_builder)
        request = request_preparer.create_request_builder(request_definition)
        assert transaction_hook_mock.audit_request.called
        assert isinstance(request, helpers.RequestBuilder)


class TestCallFactory(object):
    def test_call(self, mocker, request_definition, request_builder):
        args = ()
        kwargs = {}
        request_preparer = mocker.Mock(spec=builder.RequestPreparer)
        request_preparer.create_request_builder.return_value = request_builder
        execution_builder = mocker.Mock(spec=io.RequestExecutionBuilder)
        execution_builder.build().start.return_value = object()
        factory = builder.CallFactory(
            request_preparer, request_definition, lambda: execution_builder
        )
        assert (
            factory(*args, **kwargs)
            is execution_builder.build().start.return_value
        )
        request_definition.define_request.assert_called_with(
            request_builder, args, kwargs
        )
        request_preparer.prepare_request.assert_called_with(
            request_builder, execution_builder
        )
        execution_builder.build().start.assert_called_with(
            (request_builder.method, request_builder.url, request_builder.info)
        )


class TestBuilder(object):
    def test_init_adds_standard_converter_factory(self, uplink_builder):
        assert isinstance(
            uplink_builder.converters[-1], converters.StandardConverter
        )

    def test_client_getter(self, uplink_builder, http_client_mock):
        uplink_builder.client = http_client_mock
        assert uplink_builder.client is http_client_mock

    def test_client_setter(self, uplink_builder, http_client_mock):
        uplink_builder.client = http_client_mock
        assert uplink_builder._client is http_client_mock

    def test_base_url(self, uplink_builder):
        uplink_builder.base_url = "example"
        assert uplink_builder._base_url == "example"

    def test_add_converter_factory(
        self, uplink_builder, converter_factory_mock
    ):
        uplink_builder.converters = (converter_factory_mock,)
        factory = list(uplink_builder.converters)[0]
        assert factory == converter_factory_mock

    def test_build(self, request_definition, uplink_builder):
        call = uplink_builder.build(request_definition)
        assert isinstance(call, builder.CallFactory)


def test_build_failure(fake_service_cls):
    exception = exceptions.InvalidRequestDefinition()
    fake_service_cls.builder.build.side_effect = exception
    fake_service_cls.builder.__name__ = "builder"
    with pytest.raises(exceptions.UplinkBuilderError):
        builder.build(fake_service_cls, base_url="example.com")


def test_build(
    mocker,
    http_client_mock,
    converter_factory_mock,
    fake_service_cls,
    transaction_hook_mock,
):
    # Monkey-patch the Builder class.
    builder_cls_mock = mocker.Mock()
    builder_mock = builder.Builder()
    builder_cls_mock.return_value = builder_mock
    mocker.patch.object(builder, "Builder", builder_cls_mock)

    builder.build(
        fake_service_cls,
        base_url="example.com",
        client=http_client_mock,
        converter=converter_factory_mock,
        auth=("username", "password"),
        hook=transaction_hook_mock,
    )
    assert builder_mock.base_url == "example.com"
    assert builder_mock.client is http_client_mock
    assert list(builder_mock.converters)[0] is converter_factory_mock
    assert isinstance(builder_mock.auth, auth.BasicAuth)


def test_setting_request_method(request_definition_builder):
    # We need this test, since we wrap the request definition builders and
    # want to ensure that we unwrap on attribute access.

    class Consumer(builder.Consumer):
        request_method = request_definition_builder

    # Verify: Get request definition builder on access
    assert Consumer.request_method is request_definition_builder.copy()

    # Verify: Try again after resetting
    Consumer.request_method = request_definition_builder
    assert Consumer.request_method is request_definition_builder.copy()

    # Verify: We get callable on attribute access for an instance
    consumer = Consumer()
    assert callable(consumer.request_method)


def test_inject(mocker, fake_service_cls, transaction_hook_mock):
    # Monkey-patch the Builder class.
    builder_cls_mock = mocker.Mock()
    builder_mock = mocker.Mock(wraps=builder.Builder())
    builder_cls_mock.return_value = builder_mock
    mocker.patch.object(builder, "Builder", builder_cls_mock)

    service = builder.build(fake_service_cls)

    # Verify
    service._inject(transaction_hook_mock)
    builder_mock.add_hook.assert_called_with(transaction_hook_mock)



"""
This module implements the built-in argument annotations and their
handling classes.
"""
# Standard library imports
import collections
from collections import abc
import functools
import inspect

# Local imports
from uplink import exceptions, hooks, interfaces, utils
from uplink.converters import keys

__all__ = [
    "Path",
    "Query",
    "QueryMap",
    "Header",
    "HeaderMap",
    "Field",
    "FieldMap",
    "Part",
    "PartMap",
    "Body",
    "Url",
    "Timeout",
    "Context",
]


class ExhaustedArguments(exceptions.AnnotationError):
    message = (
        "Failed to add `%s` to method `%s`, as all arguments have "
        "been annotated."
    )

    def __init__(self, annotation, func):
        self.message = self.message % (annotation, func.__name__)


class ArgumentNotFound(exceptions.AnnotationError):
    message = "`%s` does not match any argument name of method `%s`."

    def __init__(self, name, func):
        self.message = self.message % (name, func.__name__)


class ArgumentAnnotationHandlerBuilder(interfaces.AnnotationHandlerBuilder):
    __ANNOTATION_BUILDER_KEY = "#ANNOTATION_BUILDER_KEY#"

    @classmethod
    def from_func(cls, func):
        if not hasattr(func, cls.__ANNOTATION_BUILDER_KEY):
            spec = utils.get_arg_spec(func)
            handler = cls(func, spec.args)
            setattr(func, cls.__ANNOTATION_BUILDER_KEY, handler)
            handler.set_annotations(spec.annotations)
        return getattr(func, cls.__ANNOTATION_BUILDER_KEY)

    def __init__(self, func, arguments, func_is_method=True):
        self._arguments = arguments[func_is_method:]
        self._annotations = collections.OrderedDict.fromkeys(self._arguments)
        self._defined = 0
        self._func = func
        self._argument_types = {}

    @property
    def missing_arguments(self):
        return (a for a in self._arguments if self._annotations[a] is None)

    @property
    def remaining_args_count(self):
        return len(self._arguments) - self._defined

    def set_annotations(self, annotations=None, **more_annotations):
        if annotations is not None:
            if not isinstance(annotations, abc.Mapping):
                missing = tuple(
                    a
                    for a in self.missing_arguments
                    if a not in more_annotations
                )
                annotations = dict(zip(missing, annotations))
            more_annotations.update(annotations)
        for name in more_annotations:
            self.add_annotation(more_annotations[name], name)

    @staticmethod
    def _is_annotation(annotation):
        cls = interfaces.Annotation
        return utils.is_subclass(annotation, cls) or isinstance(annotation, cls)

    def add_annotation(self, annotation, name=None, *args, **kwargs):
        if self._is_annotation(annotation):
            return self._add_annotation(annotation, name)
        else:
            self._argument_types[name] = annotation

    def _add_annotation(self, annotation, name=None):
        try:
            name = next(self.missing_arguments) if name is None else name
        except StopIteration:
            raise ExhaustedArguments(annotation, self._func)
        if name not in self._annotations:
            raise ArgumentNotFound(name, self._func)
        annotation = self._process_annotation(name, annotation)
        super(ArgumentAnnotationHandlerBuilder, self).add_annotation(annotation)
        self._defined += self._annotations[name] is None
        self._annotations[name] = annotation
        return annotation

    def _process_annotation(self, name, annotation):
        if inspect.isclass(annotation):
            annotation = annotation()
        if isinstance(annotation, TypedArgument) and annotation.type is None:
            annotation.type = self._argument_types.pop(name, None)
        if isinstance(annotation, NamedArgument) and annotation.name is None:
            annotation.name = name
        return annotation

    def is_done(self):
        return self.remaining_args_count == 0

    def copy(self):
        return self

    @property
    def _types(self):
        types = self._annotations
        return ((k, types[k]) for k in types if types[k] is not None)

    def build(self):
        return ArgumentAnnotationHandler(
            self._func, collections.OrderedDict(self._types)
        )


class ArgumentAnnotationHandler(interfaces.AnnotationHandler):
    def __init__(self, func, arguments):
        self._func = func
        self._arguments = arguments

    @property
    def annotations(self):
        return iter(self._arguments.values())

    def get_relevant_arguments(self, call_args):
        annotations = self._arguments
        return ((n, annotations[n]) for n in call_args if n in annotations)

    def handle_call(self, request_builder, args, kwargs):
        call_args = utils.get_call_args(self._func, None, *args, **kwargs)
        self.handle_call_args(request_builder, call_args)

    def handle_call_args(self, request_builder, call_args):
        # TODO: Catch Annotation errors and chain them here + provide context.
        for name, annotation in self.get_relevant_arguments(call_args):
            annotation.modify_request(request_builder, call_args[name])


class ArgumentAnnotation(interfaces.Annotation):
    _can_be_static = True

    def __call__(self, request_definition_builder):
        request_definition_builder.argument_handler_builder.add_annotation(self)
        return request_definition_builder

    def _modify_request(self, request_builder, value):  # pragma: no cover
        pass

    @property
    def type(self):  # pragma: no cover
        return None

    @property
    def converter_key(self):  # pragma: no cover
        raise NotImplementedError

    def modify_request(self, request_builder, value):
        argument_type, converter_key = self.type, self.converter_key
        converter = request_builder.get_converter(converter_key, argument_type)
        converted_value = converter(value) if converter else value
        self._modify_request(request_builder, converted_value)


class TypedArgument(ArgumentAnnotation):
    def __init__(self, type=None):
        self._type = type

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, type_):
        if self._type is None:
            self._type = type_
        else:
            raise AttributeError("Type is already set.")

    @property
    def converter_key(self):  # pragma: no cover
        raise NotImplementedError


class NamedArgument(TypedArgument):
    _can_be_static = True

    def __init__(self, name=None, type=None):
        self._arg_name = name
        super(NamedArgument, self).__init__(type)

    @property
    def name(self):
        return self._arg_name

    @name.setter
    def name(self, name):
        if self._arg_name is None:
            self._arg_name = name
        else:
            raise AttributeError("Name is already set.")

    @property
    def converter_key(self):  # pragma: no cover
        raise NotImplementedError


class EncodeNoneMixin(object):
    #: Identifies how a `None` value should be encoded in the request.
    _encode_none = None  # type: str

    def _modify_request(self, request_builder, value):  # pragma: no cover
        raise NotImplementedError

    def modify_request(self, request_builder, value):
        if value is None:
            if self._encode_none is None:
                # ignore value if it is None and shouldn't be encoded
                return
            else:
                value = self._encode_none
        super(EncodeNoneMixin, self).modify_request(request_builder, value)


class FuncDecoratorMixin(object):
    @classmethod
    def _is_static_call(cls, *args_, **kwargs):
        if super(FuncDecoratorMixin, cls)._is_static_call(*args_, **kwargs):
            return True
        try:
            is_func = inspect.isfunction(args_[0])
        except IndexError:
            return False
        else:
            return is_func and not (kwargs or args_[1:])

    def __call__(self, obj):
        if inspect.isfunction(obj):
            ArgumentAnnotationHandlerBuilder.from_func(obj).add_annotation(self)
            return obj
        else:
            return super(FuncDecoratorMixin, self).__call__(obj)

    def with_value(self, value):
        """
        Creates an object that can be used with the
        :py:class:`Session.inject` method or
        :py:class:`~uplink.inject` decorator to inject request properties
        with specific values.

        .. versionadded:: 0.4.0
        """
        auditor = functools.partial(self.modify_request, value=value)
        return hooks.RequestAuditor(auditor)


class Path(NamedArgument):
    """
    Substitution of a path variable in a `URI template
    <https://tools.ietf.org/html/rfc6570>`__.

    URI template parameters are enclosed in braces (e.g.,
    :code:`{name}`). To map an argument to a declared URI parameter, use
    the :py:class:`Path` annotation:

    .. code-block:: python

        class TodoService(object):
            @get("/todos{/id}")
            def get_todo(self, todo_id: Path("id")): pass

    Then, invoking :code:`get_todo` with a consumer instance:

    .. code-block:: python

        todo_service.get_todo(100)

    creates an HTTP request with a URL ending in :code:`/todos/100`.

    Note:
        Any unannotated function argument that shares a name with a URL path
        parameter is implicitly annotated with this class at runtime.

        For example, we could simplify the method from the previous
        example by matching the path variable and method argument names:

        .. code-block:: python

            @get("/todos{/id}")
            def get_todo(self, id): pass
    """

    @property
    def converter_key(self):
        return keys.CONVERT_TO_STRING

    def modify_request_definition(self, request_definition_builder):
        request_definition_builder.uri.add_variable(self.name)

    def _modify_request(self, request_builder, value):
        request_builder.set_url_variable({self.name: value})


class Query(FuncDecoratorMixin, EncodeNoneMixin, NamedArgument):
    """
    Set a dynamic query parameter.

    This annotation turns argument values into URL query
    parameters. You can include it as function argument
    annotation, in the format: ``<query argument>: uplink.Query``.

    If the API endpoint you are trying to query uses ``q`` as a query
    parameter, you can add ``q: uplink.Query`` to the consumer method to
    set the ``q`` search term at runtime.

    Example:
        .. code-block:: python

            @get("/search/commits")
            def search(self, search_term: Query("q")):
                \"""Search all commits with the given search term.\"""

        To specify whether or not the query parameter is already URL encoded,
        use the optional :py:obj:`encoded` argument:

        .. code-block:: python

            @get("/search/commits")
            def search(self, search_term: Query("q", encoded=True)):
                \"""Search all commits with the given search term.\"""

        To specify if and how :py:obj:`None` values should be encoded, use the
        optional :py:obj:`encode_none` argument:

        .. code-block:: python

            @get("/search/commits")
            def search(self, search_term: Query("q"),
                       search_order: Query("o", encode_none="null")):
                \"""
                Search all commits with the given search term using the
                optional search order.
                \"""

    Args:
        encoded (:obj:`bool`, optional): Specifies whether the parameter
            :py:obj:`name` and value are already URL encoded.
        encode_none (:obj:`str`, optional): Specifies an optional string with
            which :py:obj:`None` values should be encoded. If not specified,
            parameters with a value of :py:obj:`None` will not be sent.
    """

    class QueryStringEncodingError(exceptions.AnnotationError):
        message = "Failed to join encoded and unencoded query parameters."

    def __init__(self, name=None, encoded=False, type=None, encode_none=None):
        super(Query, self).__init__(name, type)
        self._encoded = encoded
        self._encode_none = encode_none

    @staticmethod
    def _update_params(info, existing, new_params, encoded):
        # TODO: Consider moving some of this to the client backend.
        if encoded:
            params = [] if existing is None else [str(existing)]
            params.extend("%s=%s" % (n, new_params[n]) for n in new_params)
            info["params"] = "&".join(params)
        else:
            info["params"].update(new_params)

    @staticmethod
    def update_params(info, new_params, encoded):
        existing = info.setdefault("params", None if encoded else dict())
        if encoded == isinstance(existing, abc.Mapping):
            raise Query.QueryStringEncodingError()
        Query._update_params(info, existing, new_params, encoded)

    @property
    def converter_key(self):
        """Converts query parameters to the request body."""
        if self._encoded:
            return keys.CONVERT_TO_STRING
        else:
            return keys.Sequence(keys.CONVERT_TO_STRING)

    def _modify_request(self, request_builder, value):
        """Updates request body with the query parameter."""
        self.update_params(
            request_builder.info, {self.name: value}, self._encoded
        )


class QueryMap(FuncDecoratorMixin, TypedArgument):
    """
    A mapping of query arguments.

    If the API you are using accepts multiple query arguments, you can
    include them all in your function method by using the format:
    ``<query argument>: uplink.QueryMap``

    Example:
        .. code-block:: python

            @get("/search/users")
            def search(self, **params: QueryMap):
                \"""Search all users.\"""

    Args:
        encoded (:obj:`bool`, optional): Specifies whether the parameter
            :py:obj:`name` and value are already URL encoded.
    """

    def __init__(self, encoded=False, type=None):
        super(QueryMap, self).__init__(type)
        self._encoded = encoded

    @property
    def converter_key(self):
        """Converts query mapping to request body."""
        if self._encoded:
            return keys.Map(keys.CONVERT_TO_STRING)
        else:
            return keys.Map(keys.Sequence(keys.CONVERT_TO_STRING))

    def _modify_request(self, request_builder, value):
        """Updates request body with the mapping of query args."""
        Query.update_params(request_builder.info, value, self._encoded)


class Header(FuncDecoratorMixin, EncodeNoneMixin, NamedArgument):
    """
    Pass a header as a method argument at runtime.

    While :py:class:`uplink.headers` attaches request headers values
    that are static across all requests sent from the decorated
    consumer method, this annotation turns a method argument into a
    dynamic request header.

    Example:
        .. code-block:: python

            @get("/user")
            def me(self, session_id: Header("Authorization")):
                \"""Get the authenticated user.\"""

        To define an optional header, use the default value of `None`:

            @get("/repositories")
            def fetch_repos(self, auth: Header("Authorization") = None):
                \"""List all public repositories.\"""

        When the argument is not used, the header will not be added
        to the request.
    """

    @property
    def converter_key(self):
        """Converts passed argument to string."""
        return keys.CONVERT_TO_STRING

    def _modify_request(self, request_builder, value):
        """Updates request header contents."""
        request_builder.info["headers"][self.name] = value


class HeaderMap(FuncDecoratorMixin, TypedArgument):
    """Pass a mapping of header fields at runtime."""

    @property
    def converter_key(self):
        """Converts every header field to string"""
        return keys.Map(keys.CONVERT_TO_STRING)

    @classmethod
    def _modify_request(cls, request_builder, value):
        """Updates request header contents."""
        request_builder.info["headers"].update(value)


class Field(NamedArgument):
    """
    Defines a form field to the request body.

    Use together with the decorator :py:class:`uplink.form_url_encoded`
    and annotate each argument accepting a form field with
    :py:class:`uplink.Field`.

    Example::
        .. code-block:: python

            @form_url_encoded
            @post("/users/edit")
            def update_user(self, first_name: Field, last_name: Field):
                \"""Update the current user.\"""
    """

    class FieldAssignmentFailed(exceptions.AnnotationError):
        """Used if the field chosen failed to be defined."""

        message = (
            "Failed to define field '%s' to request body. Another argument "
            "annotation might have overwritten the body entirely."
        )

        def __init__(self, field):
            self.message = self.message % field.name

    @property
    def converter_key(self):
        """Converts type to request body."""
        return keys.CONVERT_TO_REQUEST_BODY

    def _modify_request(self, request_builder, value):
        """Updates the request body with chosen field."""
        try:
            request_builder.info["data"][self.name] = value
        except TypeError:
            # TODO: re-raise with TypeError
            # `data` does not support item assignment
            raise self.FieldAssignmentFailed(self)


class FieldMap(TypedArgument):
    """
    Defines a mapping of form fields to the request body.

    Use together with the decorator :py:class:`uplink.form_url_encoded`
    and annotate each argument accepting a form field with
    :py:class:`uplink.FieldMap`.

    Example:
        .. code-block:: python

            @form_url_encoded
            @post("/user/edit")
            def create_post(self, **user_info: FieldMap):
                \"""Update the current user.\"""
    """

    class FieldMapUpdateFailed(exceptions.AnnotationError):
        """Use when the attempt to update the request body failed."""

        message = (
            "Failed to update request body with field map. Another argument "
            "annotation might have overwritten the body entirely."
        )

    @property
    def converter_key(self):
        """Converts type to request body."""
        return keys.Map(keys.CONVERT_TO_REQUEST_BODY)

    def _modify_request(self, request_builder, value):
        """Updates request body with chosen field mapping."""
        try:
            request_builder.info["data"].update(value)
        except AttributeError:
            # TODO: re-raise with AttributeError
            raise self.FieldMapUpdateFailed()


class Part(NamedArgument):
    """
    Marks an argument as a form part.

    Use together with the decorator :py:class:`uplink.multipart` and
    annotate each form part with :py:class:`uplink.Part`.

    Example:
        .. code-block:: python

            @multipart
            @put(/user/photo")
            def update_user(self, photo: Part, description: Part):
                \"""Upload a user profile photo.\"""
    """

    @property
    def converter_key(self):
        """Converts part to the request body."""
        return keys.CONVERT_TO_REQUEST_BODY

    def _modify_request(self, request_builder, value):
        """Updates the request body with the form part."""
        request_builder.info["files"][self.name] = value


class PartMap(TypedArgument):
    """
    A mapping of form field parts.

    Use together with the decorator :py:class:`uplink.multipart` and
    annotate each part of form parts with :py:class:`uplink.PartMap`

    Example:
        .. code-block:: python

            @multipart
            @put(/user/photo")
            def update_user(self, photo: Part, description: Part):
                \"""Upload a user profile photo.\"""
    """

    @property
    def converter_key(self):
        """Converts each part to the request body."""
        return keys.Map(keys.CONVERT_TO_REQUEST_BODY)

    def _modify_request(self, request_builder, value):
        """Updates request body to with the form parts."""
        request_builder.info["files"].update(value)


class Body(TypedArgument):
    """
    Set the request body at runtime.

    Use together with the decorator :py:class:`uplink.json`. The method
    argument value will become the request's body when annotated
    with :py:class:`uplink.Body`.

    Example:
        .. code-block:: python

            @json
            @patch(/user")
            def update_user(self, **info: Body):
                \"""Update the current user.\"""
    """

    @property
    def converter_key(self):
        """Converts request body."""
        return keys.CONVERT_TO_REQUEST_BODY

    def _modify_request(self, request_builder, value):
        """Updates request body data."""
        request_builder.info["data"] = value


# TODO: Add an integration test that uses arguments.Url
class Url(ArgumentAnnotation):
    """
    Sets a dynamic URL.

    Provides the URL at runtime as a method argument. Drop the decorator
    parameter path from :py:class:`uplink.get` and annotate the
    corresponding argument with :py:class:`uplink.Url`

    Example:
        .. code-block:: python

            @get
            def get(self, endpoint: Url):
                \"""Execute a GET requests against the given endpoint\"""
    """

    class DynamicUrlAssignmentFailed(exceptions.InvalidRequestDefinition):
        """Raised when the attempt to set dynamic url fails."""

        message = "Failed to set dynamic url annotation on `%s`. "

        def __init__(self, request_definition_builder):
            self.message = self.message % request_definition_builder.__name__

    @property
    def converter_key(self):
        """Converts url type to string."""
        return keys.CONVERT_TO_STRING

    def modify_request_definition(self, request_definition_builder):
        """Sets dynamic url."""
        try:
            request_definition_builder.uri.is_dynamic = True
        except ValueError:
            # TODO: re-raise with ValueError
            raise self.DynamicUrlAssignmentFailed(request_definition_builder)

    @classmethod
    def _modify_request(cls, request_builder, value):
        """Updates request url."""
        request_builder.relative_url = value


class Timeout(FuncDecoratorMixin, ArgumentAnnotation):
    """
    Passes a timeout as a method argument at runtime.

    While :py:class:`uplink.timeout` attaches static timeout to all requests
    sent from a consumer method, this class turns a method argument into a
    dynamic timeout value.

    Example:
        .. code-block:: python

            @get("/user/posts")
            def get_posts(self, timeout: Timeout() = 60):
                \"""Fetch all posts for the current users giving up after given
                number of seconds.\"""
    """

    @property
    def type(self):
        return float

    @property
    def converter_key(self):
        """Do not convert passed argument."""
        return keys.Identity()

    def _modify_request(self, request_builder, value):
        """Modifies request timeout."""
        request_builder.info["timeout"] = value


class Context(FuncDecoratorMixin, NamedArgument):
    """
    Defines a name-value pair that is accessible to middleware at
    runtime.

    Request middleware can leverage this annotation to give users
    control over the middleware's behavior.

    Example:
        Consider a custom decorator :obj:`@cache` (this would be a
        subclass of :class:`uplink.decorators.MethodAnnotation`):

        .. code-block:: python

            @cache(hours=3)
            @get("users/user_id")
            def get_user(self, user_id)
                \"""Retrieves a single user.\"""

        As its name suggests, the :obj:`@cache` decorator enables
        caching server responses so that, once a request is cached,
        subsequent identical requests can be served by the cache
        provider rather than adding load to the upstream service.

        Importantly, the writers of the :obj:`@cache` decorators can
        allow users to pass their own cache provider implementation
        through an argument annotated with :class:`Context <uplink.Context>`:

        .. code-block:: python

            @cache(hours=3)
            @get("users/user_id")
            def get_user(self, user_id, cache_provider: Context)
                \"""Retrieves a single user.\"""

        To add a name-value pair to the context of any request made from
        a :class:`Consumer <uplink.Consumer>` instance, you
        can use the :attr:`Consumer.session.context
        <uplink.session.Session.context>` property. Alternatively, you
        can annotate a constructor argument of a :class:`Consumer
        <uplink.Consumer>` subclass with :class:`Context
        <uplink.Context>`, as explained
        :ref:`here <annotating constructor arguments>`.
    """

    @property
    def converter_key(self):
        """Do not convert passed argument."""
        return keys.Identity()

    def _modify_request(self, request_builder, value):
        """Sets the name-value pair in the context."""
        request_builder.context[self.name] = value


class ContextMap(FuncDecoratorMixin, ArgumentAnnotation):
    """
    Defines a mapping of name-value pairs that are accessible to
    middleware at runtime.
    """

    @property
    def converter_key(self):
        """Do not convert passed argument."""
        return keys.Identity()

    def _modify_request(self, request_builder, value):
        """Updates the context with the given name-value pairs."""
        if not isinstance(value, abc.Mapping):
            raise TypeError(
                "ContextMap requires a mapping; got %s instead.", type(value)
            )
        request_builder.context.update(value)



# Standard library imports
import collections
import inspect

try:
    # Python 3.2+
    from inspect import signature
except ImportError:  # pragma: no cover
    # Python 2.7
    from inspect import getcallargs as get_call_args, getargspec as _getargspec

    def signature(_):
        raise ImportError

    def get_arg_spec(f):
        arg_spec = _getargspec(f)
        args = arg_spec.args
        if arg_spec.varargs is not None:
            args.append(arg_spec.varargs)
        if arg_spec.keywords is not None:
            args.append(arg_spec.keywords)
        return Signature(args, {}, None)


else:  # pragma: no cover

    def get_call_args(f, *args, **kwargs):
        sig = signature(f)
        arguments = sig.bind(*args, **kwargs).arguments
        # apply defaults:
        new_arguments = []
        for name, param in sig.parameters.items():
            try:
                new_arguments.append((name, arguments[name]))
            except KeyError:
                if param.default is not param.empty:
                    val = param.default
                elif param.kind is param.VAR_POSITIONAL:
                    val = ()
                elif param.kind is param.VAR_KEYWORD:
                    val = {}
                else:
                    continue
                new_arguments.append((name, val))
        return collections.OrderedDict(new_arguments)

    def get_arg_spec(f):
        sig = signature(f)
        parameters = sig.parameters
        args = []
        annotations = collections.OrderedDict()
        has_return_type = sig.return_annotation is not sig.empty
        return_type = sig.return_annotation if has_return_type else None
        for p in parameters:
            if parameters[p].annotation is not sig.empty:
                annotations[p] = parameters[p].annotation
            args.append(p)
        return Signature(args, annotations, return_type)


try:
    import urllib.parse as _urlparse
except ImportError:
    import urlparse as _urlparse


# Third-party imports
import uritemplate


urlparse = _urlparse

Signature = collections.namedtuple(
    "Signature", "args annotations return_annotation"
)

Request = collections.namedtuple("Request", "method uri info return_type")


def is_subclass(cls, class_info):
    return inspect.isclass(cls) and issubclass(cls, class_info)


def no_op(*_, **__):
    pass


class URIBuilder(object):
    @staticmethod
    def variables(uri):
        try:
            return uritemplate.URITemplate(uri).variable_names
        except TypeError:
            return set()

    def __init__(self, uri):
        self._uri = uritemplate.URITemplate(uri or "")

    def set_variable(self, var_dict=None, **kwargs):
        self._uri = self._uri.partial(var_dict, **kwargs)

    def remaining_variables(self):
        return self._uri.variable_names

    def build(self):
        return self._uri.expand()



# Local imports
from uplink import retry
from uplink.retry import backoff, stop, when


def test_jittered_backoff():
    iterator = backoff.jittered()()

    first = next(iterator)
    assert 0 <= first <= 1

    second = next(iterator)
    assert 0 <= second <= 2

    third = next(iterator)
    assert 0 <= third <= 4

    fourth = next(iterator)
    assert 0 <= fourth <= 8

    fifth = next(iterator)
    assert 0 <= fifth <= 16


def test_exponential_backoff_minimum():
    iterator = backoff.exponential(base=2, minimum=8)()
    assert next(iterator) == 8
    assert next(iterator) == 16


def test_fixed_backoff():
    iterator = backoff.fixed(10)()
    assert next(iterator) == 10
    assert next(iterator) == 10
    assert next(iterator) == 10


def test_retry_stop_default():
    decorator = retry()
    assert stop.NEVER == decorator._stop

    stop_gen = decorator._stop()

    next(stop_gen)
    assert stop_gen.send(1) is False

    next(stop_gen)
    assert stop_gen.send(5) is False

    next(stop_gen)
    assert stop_gen.send(10) is False

    next(stop_gen)
    assert stop_gen.send(10.1) is False


def test_stop_or():
    stop_gen = (stop.after_delay(2) | stop.after_attempt(3))()

    next(stop_gen)
    assert stop_gen.send(3) is True

    next(stop_gen)
    assert stop_gen.send(1) is False

    next(stop_gen)
    assert stop_gen.send(5) is True

    next(stop_gen)
    assert stop_gen.send(1) is True


def test_stop_or_with_none():
    stop1 = stop.after_delay(2)
    assert stop1 is (stop1 | None)


def test_retry_custom_stop():
    def custom_stop(*_):
        return True

    decorator = retry(stop=custom_stop)
    assert decorator._stop == custom_stop


def test_retry_backoff():
    def custom_backoff(*_):
        return True

    decorator = retry(backoff=custom_backoff)
    assert decorator._backoff == custom_backoff


def test_retry_decorator_exposes_submodules_as_properties():
    assert retry.backoff is backoff
    assert retry.stop is stop
    assert retry.when is when


def test_stop_after_delay():
    stop_gen = stop.after_delay(10)()

    # Start generator
    next(stop_gen)
    assert stop_gen.send(1) is False

    next(stop_gen)
    assert stop_gen.send(5) is False

    next(stop_gen)
    assert stop_gen.send(10) is False

    next(stop_gen)
    assert stop_gen.send(10.1) is True


def test_when_status(mocker, request_builder):
    # Setup
    response = mocker.Mock()
    response.status_code = 401

    # Verify: Returns self on call
    predicate = when.status(401)
    assert predicate(request_builder) is predicate

    # Verify: Encountered expected status
    assert when.status(401).should_retry_after_response(response) is True

    # Verify: Encountered unexpected status
    assert when.status(200).should_retry_after_response(response) is False

    # Verify: Should return false for exceptions
    assert (
        when.status(401).should_retry_after_exception(
            Exception, Exception(), None
        )
        is False
    )


def test_when_status_5xx(mocker, request_builder):
    # Setup
    response = mocker.Mock()
    response.status_code = 501

    # Verify: Returns self on call
    predicate = when.status_5xx()
    assert predicate(request_builder) is predicate

    # Verify: Encountered bad request status
    assert when.status_5xx().should_retry_after_response(response) is True

    # Verify: Encountered successful request status
    response.status_code = 200
    assert when.status_5xx().should_retry_after_response(response) is False

    # Verify: Should return false for exceptions
    assert (
        when.status_5xx().should_retry_after_exception(
            Exception, Exception(), None
        )
        is False
    )


def test_when_or_operator_with_response(mocker):
    # Setup
    response = mocker.Mock()
    response.status_code = 401
    predicate = when.status(401) | when.status(405)

    # Verify: when left predicate matches
    assert predicate.should_retry_after_response(response) is True

    # Verify: when right predicate matches
    response.status_code = 405
    assert predicate.should_retry_after_response(response) is True

    # Verify: when neither matches
    response.status_code = 200
    assert predicate.should_retry_after_response(response) is False


def test_when_or_operator_with_exception(mocker, request_builder):
    # Setup
    response = mocker.Mock()
    response.status_code = 401
    predicate = when.status(401) | when.raises(retry.BASE_CLIENT_EXCEPTION)

    # Verify: Calls __call__ for both predicates
    request_builder.client.exceptions.BaseClientException = RuntimeError
    predicate = predicate(request_builder)

    # Verify: when left predicate matches
    assert predicate.should_retry_after_response(response) is True

    # Verify: when right predicate matches
    assert (
        predicate.should_retry_after_exception(
            RuntimeError, RuntimeError(), None
        )
        is True
    )

    # Verify: when neither matches
    assert (
        predicate.should_retry_after_exception(Exception, Exception(), None)
        is False
    )


class TestClientExceptionProxies(object):
    @staticmethod
    def _get_exception(proxy, request_builder):
        return proxy(request_builder.client.exceptions)

    def test_basic_client_exception(self, request_builder):
        exc = self._get_exception(retry.BASE_CLIENT_EXCEPTION, request_builder)
        assert request_builder.client.exceptions.BaseClientException == exc

    def test_connection_error(self, request_builder):
        exc = self._get_exception(retry.CONNECTION_ERROR, request_builder)
        assert request_builder.client.exceptions.BaseClientException == exc

    def test_connection_timeout(self, request_builder):
        exc = self._get_exception(retry.CONNECTION_TIMEOUT, request_builder)
        assert request_builder.client.exceptions.BaseClientException == exc

    def test_server_timeout(self, request_builder):
        exc = self._get_exception(retry.SERVER_TIMEOUT, request_builder)
        assert request_builder.client.exceptions.BaseClientException == exc

    def test_ssl_error(self, request_builder):
        exc = self._get_exception(retry.SSL_ERROR, request_builder)
        assert request_builder.client.exceptions.BaseClientException == exc



# Local imports
from uplink.clients import interfaces

#: Provide this flag to :py:func:`get_client` to create an instance of
#: the default HTTP client adapter.
DEFAULT_CLIENT = None

# (default client, handlers)
_registrar = [None, []]


def handler(func):
    """Registers :py:obj:`func` as a handler."""
    # TODO: support handler prioritization?
    _registrar[1].append(func)


def handle_client_key(key):
    # Try handlers
    for func in _registrar[1]:
        client = func(key)
        if isinstance(client, interfaces.HttpClientAdapter):
            return client


def set_default_client(client):
    _registrar[0] = client


def get_default_client():
    default_client = _registrar[0]
    if callable(default_client):
        return default_client()
    else:
        return default_client


def get_client(client=DEFAULT_CLIENT):
    if client is DEFAULT_CLIENT:
        client = get_default_client()

    if isinstance(client, interfaces.HttpClientAdapter):
        return client
    else:
        # Try handlers
        return handle_client_key(client)



# Local imports
from uplink.clients import exceptions, io


class HttpClientAdapter(io.Client):
    """An adapter of an HTTP client library."""

    __exceptions = exceptions.Exceptions()

    def io(self):
        """Returns the execution strategy for this client."""
        raise NotImplementedError

    @property
    def exceptions(self):
        """
        uplink.clients.exceptions.Exceptions: An enum of standard HTTP
        client errors that have been mapped to client specific
        exceptions.
        """
        return self.__exceptions

    def send(self, request):
        raise NotImplementedError

    def apply_callback(self, callback, response):
        raise NotImplementedError



# Standard library imports
import collections

# Local imports.
import uplink

# Constants
BASE_URL = "https://api.github.com/"

# Schemas
User = collections.namedtuple("User", "id name")
Repo = collections.namedtuple("Repo", "owner name")

# Converters


@uplink.loads(User)
def user_reader(cls, response):
    return cls(**response.json())


@uplink.loads.from_json(Repo)
def repo_json_reader(cls, json):
    return cls(**json)


@uplink.dumps.to_json(Repo)
def repo_json_writer(_, repo):
    return {"owner": repo.owner, "name": repo.name}


# Service


class GitHub(uplink.Consumer):
    @uplink.returns(User)
    @uplink.get("/users/{user}")
    def get_user(self, user):
        pass

    @uplink.returns.from_json(type=Repo)
    @uplink.get("/users/{user}/repos/{repo}")
    def get_repo(self, user, repo):
        pass

    @uplink.returns.from_json(key="data")
    @uplink.returns.schema(uplink.types.List[Repo])
    @uplink.get("/users/{user}/repos")
    def get_repos(self, user):
        pass

    @uplink.json
    @uplink.post("/users/{user}/repos", args={"repo": uplink.Body(Repo)})
    def create_repo(self, user, repo):
        pass


# Tests


def test_returns_with_type(mock_client, mock_response):
    # Setup
    mock_response.with_json({"id": 123, "name": "prkumar"})
    mock_client.with_response(mock_response)
    github = GitHub(
        base_url=BASE_URL, client=mock_client, converters=user_reader
    )

    # Run
    user = github.get_user("prkumar")

    # Verify
    assert User(id=123, name="prkumar") == user


def test_returns_json_with_type(mock_client, mock_response):
    # Setup
    mock_response.with_json({"owner": "prkumar", "name": "uplink"})
    mock_client.with_response(mock_response)
    github = GitHub(
        base_url=BASE_URL, client=mock_client, converters=repo_json_reader
    )

    # Run
    repo = github.get_repo("prkumar", "uplink")

    # Verify
    assert Repo(owner="prkumar", name="uplink") == repo


def test_returns_json_with_list(mock_client, mock_response):
    # Setup
    mock_response.with_json(
        {
            "data": [
                {"owner": "prkumar", "name": "uplink"},
                {"owner": "prkumar", "name": "uplink-protobuf"},
            ],
            "errors": [],
        }
    )
    mock_client.with_response(mock_response)
    github = GitHub(
        base_url=BASE_URL, client=mock_client, converters=repo_json_reader
    )

    # Run
    repo = github.get_repos("prkumar")

    # Verify
    assert [
        Repo(owner="prkumar", name="uplink"),
        Repo(owner="prkumar", name="uplink-protobuf"),
    ] == repo


def test_post_json(mock_client):
    # Setup
    github = GitHub(
        base_url=BASE_URL, client=mock_client, converters=repo_json_writer
    )
    github.create_repo("prkumar", Repo(owner="prkumar", name="uplink"))
    request = mock_client.history[0]
    assert request.json == {"owner": "prkumar", "name": "uplink"}



# Local imports
from uplink import converters

__all__ = ["List", "Dict"]

List = converters.TypingConverter.List
"""
A proxy for :py:class:`typing.List` that is safe to use in type
hints with Python 3.4 and below.

.. code-block:: python

    @get("/users")
    def get_users(self) -> types.List[str]:
        \"""Fetches all users\"""
"""

Dict = converters.TypingConverter.Dict
"""
A proxy for :py:class:`typing.Dict` that is safe to use in type
hints with Python 3.4 and below.

.. code-block:: python

    @returns.from_json
    @get("/users")
    def get_users(self) -> types.Dict[str, str]:
        \"""Fetches all users\"""
"""



# Local imports
from uplink.clients.io import interfaces


class _BaseState(interfaces.RequestState):
    def __init__(self, request):
        self._request = request

    def prepare(self, request):
        return BeforeRequest(request)

    def send(self, request):
        return SendRequest(request)

    def fail(self, exc_type, exc_val, exc_tb):
        # Terminal
        return Fail(self._request, exc_type, exc_val, exc_tb)

    def finish(self, response):
        # Terminal
        return Finish(self._request, response)

    def sleep(self, duration):
        return Sleep(self._request, duration)

    def execute(self, execution):  # pragma: no cover
        raise NotImplementedError

    @property
    def request(self):
        return self._request


class BeforeRequest(_BaseState):
    def execute(self, execution):
        return execution.before_request(self._request)

    def __eq__(self, other):
        return (
            isinstance(other, BeforeRequest) and self.request == other.request
        )


class Sleep(interfaces.RequestState):
    class _Callback(interfaces.SleepCallback):
        def __init__(self, execution, request):
            self._context = execution
            self._request = request

        def on_success(self):
            self._context.state = BeforeRequest(self._request)
            return self._context.execute()

        def on_failure(self, exc_type, exc_val, exc_tb):
            self._context.state = AfterException(
                self._request, exc_type, exc_val, exc_tb
            )
            return self._context.execute()

    def __init__(self, request, duration):
        self._request = request
        self._duration = duration

    def execute(self, execution):
        return execution.sleep(
            self._duration, self._Callback(execution, self._request)
        )

    @property
    def request(self):
        return self._request

    @property
    def duration(self):
        return self._duration

    def __eq__(self, other):
        return (
            isinstance(other, Sleep)
            and self.request == other.request
            and self.duration == other.duration
        )


class SendRequest(interfaces.RequestState):
    def __init__(self, request):
        self._request = request

    class SendCallback(interfaces.InvokeCallback):
        def __init__(self, execution, request):
            self._context = execution
            self._request = request

        def on_success(self, response):
            self._context.state = AfterResponse(self._request, response)
            return self._context.execute()

        def on_failure(self, exc_type, exc_val, exc_tb):
            self._context.state = AfterException(
                self._request, exc_type, exc_val, exc_tb
            )
            return self._context.execute()

    def execute(self, execution):
        return execution.send(
            self._request, self.SendCallback(execution, self._request)
        )

    @property
    def request(self):
        return self._request

    def __eq__(self, other):
        return isinstance(other, SendRequest) and self.request == other.request


class AfterResponse(_BaseState):
    def __init__(self, request, response):
        super(AfterResponse, self).__init__(request)
        self._response = response

    def execute(self, execution):
        return execution.after_response(self._request, self._response)

    @property
    def response(self):
        return self._response

    def __eq__(self, other):
        return (
            isinstance(other, AfterResponse)
            and self.request == other.request
            and self.response == other.response
        )


class AfterException(_BaseState):
    def __init__(self, request, exc_type, exc_val, exc_tb):
        super(AfterException, self).__init__(request)
        self._exc_type = exc_type
        self._exc_val = exc_val
        self._exc_tb = exc_tb

    @property
    def exc_type(self):
        return self._exc_type

    @property
    def exc_val(self):
        return self._exc_val

    @property
    def exc_tb(self):
        return self._exc_tb

    def execute(self, execution):
        return execution.after_exception(
            self._request, self._exc_type, self._exc_val, self._exc_tb
        )

    def __eq__(self, other):
        return (
            isinstance(other, AfterException)
            and self.request == other.request
            and self.exc_type == other.exc_type
            and self.exc_val == other.exc_val
            and self.exc_tb == other.exc_tb
        )


class TerminalState(interfaces.RequestState):
    def __init__(self, request):
        self._request = request

    @property
    def request(self):
        return self._request

    def execute(self, execution):  # pragma: no cover
        raise NotImplementedError


class Fail(TerminalState):
    def __init__(self, request, exc_type, exc_val, exc_tb):
        super(Fail, self).__init__(request)
        self._exc_type = exc_type
        self._exc_val = exc_val
        self._exc_tb = exc_tb

    def execute(self, execution):
        return execution.fail(self._exc_type, self._exc_val, self._exc_tb)

    @property
    def exc_type(self):
        return self._exc_type

    @property
    def exc_val(self):
        return self._exc_val

    @property
    def exc_tb(self):
        return self._exc_tb

    def __eq__(self, other):
        return (
            isinstance(other, Fail)
            and self.request == other.request
            and self.exc_type == other.exc_type
            and self.exc_val == other.exc_val
            and self.exc_tb == other.exc_tb
        )


class Finish(TerminalState):
    def __init__(self, request, response):
        super(Finish, self).__init__(request)
        self._response = response

    def execute(self, execution):
        return execution.finish(self._response)

    @property
    def response(self):
        return self._response

    def __eq__(self, other):
        return (
            isinstance(other, Finish)
            and self.request == other.request
            and self.response == other.response
        )



# Standard library imports
import collections
import inspect

# Local imports
from uplink.converters import interfaces


class Register(object):
    def __init__(self):
        self._register = collections.deque()

    def register_converter_factory(self, proxy):
        factory = proxy() if inspect.isclass(proxy) else proxy
        if not isinstance(factory, interfaces.Factory):
            raise TypeError(
                "Failed to register '%s' as a converter factory: it is not an "
                "instance of '%s'." % (factory, interfaces.Factory)
            )
        self._register.appendleft(factory)
        return proxy

    def get_converter_factories(self):
        return tuple(self._register)


_registry = Register()

register_default_converter_factory = _registry.register_converter_factory
get_default_converter_factories = _registry.get_converter_factories



# Standard library imports
import collections
from collections import abc
import functools

# Local imports
from uplink.converters import interfaces, register_default_converter_factory

__all__ = ["TypingConverter", "ListConverter", "DictConverter"]


class BaseTypeConverter(object):
    Builder = collections.namedtuple("Builder", "build")

    @classmethod
    def freeze(cls, *args, **kwargs):
        return cls.Builder(functools.partial(cls, *args, **kwargs))


class ListConverter(BaseTypeConverter, interfaces.Converter):
    def __init__(self, elem_type):
        self._elem_type = elem_type
        self._elem_converter = None

    def set_chain(self, chain):
        self._elem_converter = chain(self._elem_type) or self._elem_type

    def convert(self, value):
        if isinstance(value, abc.Sequence):
            return list(map(self._elem_converter, value))
        else:
            # TODO: Handle the case where the value is not an sequence.
            return [self._elem_converter(value)]


class DictConverter(BaseTypeConverter, interfaces.Converter):
    def __init__(self, key_type, value_type):
        self._key_type = key_type
        self._value_type = value_type
        self._key_converter = None
        self._value_converter = None

    def set_chain(self, chain):
        self._key_converter = chain(self._key_type) or self._key_type
        self._value_converter = chain(self._value_type) or self._value_type

    def convert(self, value):
        if isinstance(value, abc.Mapping):
            key_c, val_c = self._key_converter, self._value_converter
            return dict((key_c(k), val_c(value[k])) for k in value)
        else:
            # TODO: Handle the case where the value is not a mapping.
            return self._value_converter(value)


class _TypeProxy(object):
    def __init__(self, func):
        self._func = func

    def __getitem__(self, item):
        items = item if isinstance(item, tuple) else (item,)
        return self._func(*items)


def _get_types(try_typing=True):
    if TypingConverter.typing and try_typing:
        return TypingConverter.typing.List, TypingConverter.typing.Dict
    else:
        return (
            _TypeProxy(ListConverter.freeze),
            _TypeProxy(DictConverter.freeze),
        )


@register_default_converter_factory
class TypingConverter(interfaces.Factory):
    """
    .. versionadded: v0.5.0

    An adapter that serializes and deserializes collection types from
    the :py:mod:`typing` module, such as :py:class:`typing.List`.

    Inner types of a collection are recursively resolved, using other
    available converters if necessary. For instance, when resolving the
    type hint :py:attr:`typing.Sequence[UserSchema]`, where
    :py:attr:`UserSchema` is a custom :py:class:`marshmallow.Schema`
    subclass, the converter will resolve the inner type using
    :py:class:`uplink.converters.MarshmallowConverter`.

    .. code-block:: python

        @get("/users")
        def get_users(self) -> typing.Sequence[UserSchema]:
            '''Fetch all users.'''

    Note:

        The :py:mod:`typing` module is available in the standard library
        starting from Python 3.5. For earlier versions of Python, there
        is a port of the module available on PyPI.

        However, you can utilize this converter without the
        :py:mod:`typing` module by using one of the proxies defined by
        :py:class:`uplink.returns` (e.g., :py:obj:`uplink.types.List`).
    """

    try:
        import typing
    except ImportError:  # pragma: no cover
        typing = None

    def _check_typing(self, t):
        has_origin = hasattr(t, "__origin__")
        has_args = hasattr(t, "__args__")
        return self.typing and has_origin and has_args

    def _base_converter(self, type_):
        if isinstance(type_, BaseTypeConverter.Builder):
            return type_.build()
        elif self._check_typing(type_):
            if issubclass(type_.__origin__, self.typing.Sequence):
                return ListConverter(*type_.__args__)
            elif issubclass(type_.__origin__, self.typing.Mapping):
                return DictConverter(*type_.__args__)

    def create_response_body_converter(self, type_, *args, **kwargs):
        return self._base_converter(type_)

    def create_request_body_converter(self, type_, *args, **kwargs):
        return self._base_converter(type_)


TypingConverter.List, TypingConverter.Dict = _get_types()



# Third-party imports
import pytest

# Local imports
from uplink import hooks


class TestResponseHandler(object):
    def test_handle_response(self, mocker):
        converter = mocker.Mock()
        input_value = "converted"
        converter.return_value = input_value
        rc = hooks.ResponseHandler(converter)
        assert rc.handle_response(None) is input_value


class TestRequestAuditor(object):
    def test_audit_request(self, mocker):
        auditor = mocker.stub()
        ra = hooks.RequestAuditor(auditor)
        ra.audit_request("consumer", "request")
        auditor.assert_called_with("request")


class TestExceptionHandler(object):
    def test_handle_exception_masked(self, mocker):
        handler = mocker.stub()
        eh = hooks.ExceptionHandler(handler)
        eh.handle_exception("consumer", "exc_type", "exc_val", "exc_tb")
        handler.assert_called_with("exc_type", "exc_val", "exc_tb")


class TestTransactionHookChain(object):
    def test_delegate_audit_request(self, transaction_hook_mock):
        chain = hooks.TransactionHookChain(transaction_hook_mock)
        chain.audit_request("consumer", "request")
        transaction_hook_mock.audit_request.assert_called_with(
            "consumer", "request"
        )

    def test_delegate_handle_response(self, transaction_hook_mock):
        chain = hooks.TransactionHookChain(transaction_hook_mock)
        chain.handle_response("consumer", {})
        transaction_hook_mock.handle_response.assert_called_with("consumer", {})

    def test_delegate_handle_response_multiple(self, mocker):
        # Include one hook that can't handle responses
        mock_response_handler = mocker.stub()
        mock_request_auditor = mocker.stub()

        chain = hooks.TransactionHookChain(
            hooks.RequestAuditor(mock_request_auditor),
            hooks.ResponseHandler(mock_response_handler),
            hooks.ResponseHandler(mock_response_handler),
        )
        chain.handle_response("consumer", {})
        assert mock_response_handler.call_count == 2
        assert mock_request_auditor.call_count == 0

    def test_delegate_handle_exception(self, transaction_hook_mock):
        class CustomException(Exception):
            pass

        err = CustomException()
        chain = hooks.TransactionHookChain(transaction_hook_mock)

        with pytest.raises(CustomException):
            chain.handle_exception(None, CustomException, err, None)

        transaction_hook_mock.handle_exception.assert_called_with(
            None, CustomException, err, None
        )



# Local imports
from uplink import decorators
from uplink.clients.io import RequestTemplate, transitions
from uplink.retry import (
    when as when_mod,
    stop as stop_mod,
    backoff as backoff_mod,
)
from uplink.retry._helpers import ClientExceptionProxy

__all__ = ["retry"]


class RetryTemplate(RequestTemplate):
    def __init__(self, backoff, retry_condition):
        self._backoff = backoff
        self._backoff_iterator = None
        self._condition = retry_condition
        self._reset()

    def _next_delay(self):
        try:
            delay = next(self._backoff_iterator)
        except StopIteration:
            # Fallback to the default behavior
            pass
        else:
            return transitions.sleep(delay)

    def _reset(self):
        self._backoff_iterator = self._backoff()

    def after_response(self, request, response):
        if self._condition.should_retry_after_response(response):
            return self._next_delay()
        else:
            self._reset()

    def after_exception(self, request, exc_type, exc_val, exc_tb):
        if self._condition.should_retry_after_exception(
            exc_type, exc_val, exc_tb
        ):
            return self._next_delay()
        else:
            self._reset()


# noinspection PyPep8Naming
class retry(decorators.MethodAnnotation):
    """
    A decorator that adds retry support to a consumer method or to an
    entire consumer.

    Unless you specify the ``when`` or ``on_exception`` argument, all
    failed requests that raise an exception are retried.

    Unless you specify the ``max_attempts`` or ``stop`` argument, this
    decorator continues retrying until the server returns a response.

    Unless you specify the ``backoff`` argument, this decorator uses
    `capped exponential backoff and jitter <https://amzn.to/2xc2nK2>`_,
    which should benefit performance with remote services under high
    contention.

    .. note::

        Response and error handlers (see :ref:`here <custom response
        handler>`) are invoked after the retry condition breaks or all
        retry attempts are exhausted, whatever comes first. These
        handlers will receive the first response/exception that triggers
        the retry's ``stop`` condition or doesn't match its ``when``
        filter.

        In other words, responses or exceptions that match
        the retry condition (e.g., retry when status code is 5xx) are
        not subject to response or error handlers as long as the request
        doesn't break the retry's stop condition (e.g., stop retrying
        after 5 attempts).

    Args:
        when (optional): A predicate that determines when a retry
            should be attempted.
        max_attempts (int, optional): The number of retries to attempt.
            If not specified, requests are retried continuously until
            a response is rendered.
        on_exception (:class:`Exception`, optional): The exception type
            that should prompt a retry attempt.
        stop (:obj:`callable`, optional): A function that creates
            predicates that decide when to stop retrying a request.
        backoff (:obj:`callable`, optional): A function that creates
            an iterator over the ordered sequence of timeouts between
            retries. If not specified, exponential backoff is used.
    """

    _DEFAULT_PREDICATE = when_mod.raises(Exception)

    def __init__(
        self,
        when=None,
        max_attempts=None,
        on_exception=None,
        stop=None,
        backoff=None,
    ):
        if stop is not None:
            self._stop = stop
        elif max_attempts is not None:
            self._stop = stop_mod.after_attempt(max_attempts)
        else:
            self._stop = stop_mod.NEVER

        self._predicate = when

        if on_exception is not None:
            self._predicate = when_mod.raises(on_exception) | self._predicate

        if self._predicate is None:
            self._predicate = self._DEFAULT_PREDICATE

        self._backoff = backoff_mod.jittered() if backoff is None else backoff

    BASE_CLIENT_EXCEPTION = ClientExceptionProxy(
        lambda ex: ex.BaseClientException
    )
    CONNECTION_ERROR = ClientExceptionProxy(lambda ex: ex.ConnectionError)
    CONNECTION_TIMEOUT = ClientExceptionProxy(lambda ex: ex.ConnectionTimeout)
    SERVER_TIMEOUT = ClientExceptionProxy(lambda ex: ex.ServerTimeout)
    SSL_ERROR = ClientExceptionProxy(lambda ex: ex.SSLError)

    def modify_request(self, request_builder):
        request_builder.add_request_template(
            self._create_template(request_builder)
        )

    def _create_template(self, request_builder):
        return RetryTemplate(
            self._backoff_iterator, self._predicate(request_builder)
        )

    def _backoff_iterator(self):
        stop_gen = self._stop()
        for delay in self._backoff():
            next(stop_gen)
            if stop_gen.send(delay):
                break
            yield delay

    stop = stop_mod
    backoff = backoff_mod
    when = when_mod



# Third-party imports.
import pytest

# Local imports.
import uplink
from uplink.ratelimit import RateLimitExceeded, now

# Constants
BASE_URL = "https://api.github.com/"
DIFFERENT_BASE_URL = "https://hostedgithub.com/"


class CustomRateLimitException(RuntimeError):
    pass


class GitHub(uplink.Consumer):
    @uplink.ratelimit(
        calls=1, period=10, raise_on_limit=True, group_by=None
    )  # 1 call every 10 seconds
    @uplink.get("users/{user}")
    def get_user(self, user):
        pass

    @uplink.ratelimit(calls=1, period=1, raise_on_limit=False)
    @uplink.get("repos/{user}/{repo}")
    def get_repo(self, user, repo):
        pass

    @uplink.ratelimit(
        calls=1, period=10, raise_on_limit=CustomRateLimitException
    )
    @uplink.get("repos/{user}/{repo}/comments/{comment}")
    def get_comment(self, user, repo, comment):
        pass

    @uplink.ratelimit(calls=1, period=10, raise_on_limit=True, group_by=None)
    @uplink.get("repos/{user}/{repo}/issues")
    def get_issues(self, user, repo):
        pass

    @uplink.ratelimit(calls=1, period=10, raise_on_limit=True)
    @uplink.get("repos/{user}/{repo}/issues/{issue}")
    def get_issue(self, user, repo, issue):
        pass


# Tests


def test_limit_exceeded_by_1(mock_client):
    # Setup
    github = GitHub(base_url=BASE_URL, client=mock_client)

    # Run
    github.get_user("prkumar")

    with pytest.raises(RateLimitExceeded):
        github.get_user("prkumar")


def test_limit_exceeded_by_1_with_custom_exception(mock_client):
    # Setup
    github = GitHub(base_url=BASE_URL, client=mock_client)

    # Run
    github.get_comment("prkumar", "uplink", "1")

    with pytest.raises(CustomRateLimitException):
        github.get_comment("prkumar", "uplink", "1")


def test_exceeded_limit_wait(mock_client):
    # Setup
    github = GitHub(base_url=BASE_URL, client=mock_client)

    # Run
    start = now()
    github.get_repo("prkumar", "uplink")
    github.get_repo("prkumar", "uplink")
    elapsed = now() - start

    assert elapsed >= 1


def test_limit_with_group_by_None(mock_client):
    # Setup: consumers pointing to separate hosts
    github1 = GitHub(base_url=BASE_URL, client=mock_client)
    github2 = GitHub(base_url=DIFFERENT_BASE_URL, client=mock_client)

    # Run
    github1.get_issues("prkumar", "uplink")

    # Verify: the rate limit should be applied globally for all instances
    with pytest.raises(RateLimitExceeded):
        github2.get_issues("prkumar", "uplink")


def test_limit_with_group_by_host_and_port(mock_client):
    # Setup: consumers pointing to separate hosts
    github1 = GitHub(base_url=BASE_URL, client=mock_client)
    github2 = GitHub(base_url=DIFFERENT_BASE_URL, client=mock_client)

    # Run
    github1.get_issue("prkumar", "uplink", "issue1")

    # Verify: the rate limit should be applied separately by host-port,
    # so this request should be fine.
    github2.get_issue("prkumar", "uplink", "issue2")
